"""Charging point representation.

A charging point links the charging infrastructure to the vehicles to be charged.
It is limited by its power limits where as the lower limit has to be understood as at least the
lower limit or zero.

"""
from elvis.utility.elvis_general import floor
from elvis.infrastructure_node import InfrastructureNode
from elvis.charging_station import ChargingStation


class ChargingPoint(InfrastructureNode):
    """Represents a point of connection between :obj: `charging_point.ChargingStation` and
    vehicles."""

    counter = 1

    def __init__(self, min_power, max_power, parent):
        """Create a charging point given all parameters.

        Args:
            min_power: (float): Min charging power that has to be assigned if not 0.
            max_power: (float): Max charging power that can be assigned.
            parent: (:obj: `charging_point.ChargingStation`): Charging station the charging
            point belongs to.
            """
        # A charging point must always be connected to a charging station
        assert isinstance(parent, ChargingStation)
        identification = 'cp ' + str(ChargingPoint.counter)
        ChargingPoint.counter += 1

        # set min and max power
        super().__init__(identification, min_power, max_power, parent=parent)

        self.connected_vehicle = None

    def __str__(self):
        printout = str(self.id)
        return printout

    def get_leaving_time(self):
        """Make sure a vehicle is connected and return its leaving_time."""
        assert self.connected_vehicle is not None

        return self.connected_vehicle['leaving_time']

    def connect_vehicle(self, event):
        """Assign dict of charging event as connected vehicle.

        Args:
            event: (:obj: `charging_event.ChargingEvent`): Event of car arrival."""
        self.connected_vehicle = event.to_dict(deep=False)

    def disconnect_vehicle(self):
        """Set field connected_vehicle to None so charging point is available for
            vehicle connection."""
        self.connected_vehicle = None

    def charge_vehicle(self, power, resolution):
        """Charges the vehicle according to assigned power, capacity and efficiencies.
        TODO: add efficiencies"""
        vehicle_type = self.connected_vehicle['vehicle_type']
        battery = vehicle_type.battery
        hours = resolution.total_seconds()/3600
        # TODO: This should be calculated by chaining some class methods from battery, vehicle type,
        # hardware. Each of them have different ways of converting/transporting power with their
        # specific losses
        delta = power * hours / battery.capacity

        self.connected_vehicle['soc'] += delta

    def max_hardware_power(self):
        """Calculate dependent on currently connected car the maximum power possible. Solely based
            on the charging point and battery boundaries no charging station boundaries considered.
            A further check whether this power exceeds the charging station boundaries has to be done
            separately. At the same time it is not checked whether the soc_target is met.
            TODO: Go up the tree."""

        if self.connected_vehicle is not None:
            vehicle = self.connected_vehicle
            vehicle_type = vehicle['vehicle_type']
            soc = vehicle['soc']
            battery = vehicle_type.battery

            max_power = min(self.max_power, battery.max_power_possible(soc))
            max_power = floor(max_power)
            return max_power
        return 0

    def min_hardware_power(self):
        """Calculate dependent on currently connected car the minimum power if not 0 needed
            to start charging.
            #TODO: Go up the tree."""

        if self.connected_vehicle is not None:
            vehicle = self.connected_vehicle
            vehicle_type = vehicle['vehicle_type']
            soc = vehicle['soc']
            battery = vehicle_type.battery

            min_power = max(self.min_power, battery.min_power_possible(soc))
            return min_power
        return 0

    def power_to_charge_target(self, timedelta, soc_target):
        """Calculate the average power needed in a time period to charge the battery of the
            connected vehicle to a given soc target.

        Args:
            timedelta: (:obj: `datetime.timedelta`): Length of the timeperiod.
            soc_target: (float): SOC target to be met.

        Returns:
            power_to_target: (float): Power needed to reach the given SOC target.
            If soc_target > current_soc: return 0."""

        if not self.check_soc(soc_target):
            # TODO: Raise SOC out of boundaries error.
            print('ERROR: SOC exceeded boundaries. SOC: ', soc_target)
            pass

        vehicle = self.connected_vehicle
        current_soc = self.connected_vehicle['soc']

        # if the battery is already charged further than the given SOC no power is needed
        if current_soc > soc_target:
            return 0

        vehicle_type = vehicle['vehicle_type']
        battery = vehicle_type.battery
        battery_capacity = battery.capacity
        timedelta_hours = timedelta.total_seconds() / 3600

        power_to_target = (soc_target - current_soc) * battery_capacity / timedelta_hours

        return power_to_target

    @staticmethod
    def check_soc(soc):
        """Make sure soc is in between [0, 1]"""
        if 0 <= soc <= 1.0:
            return True
        return False



"""Super class for every node in the infrastructure network.
TODO: Add node busbar."""

import networkx as nx
import matplotlib.pyplot as plt
from math import floor
from elvis.battery import StationaryBattery


class InfrastructureNode:
    """Super class of all nodes (transformer, charging station, charging point)."""
    def __init__(self, identification, min_power, max_power,
                 parent=None):
        """Create an InfrastructureNode given all parameters.

        Args:
            identification: (str): ID of Node. Node type (transformer, charging station, charging
                point) followed by "_" and an increasing integer.
            max_power: (float): Maximum power allowed to go through node.
            min_power: (float): Minimum power allowed to go through node.
            parent: (obj): Depending on position in graph the obj differs:
                transformer         -> parent = None
                charging station    -> parent = :obj: `infrastructure_node.Transformer`
                charging point      -> parent = :obj: `charging_point.ChargingPoint`
        Raises:
            AssertionError:
                - If min_power and max_power are not numerical (int or float).
                - If parent is not an instance of :obj: `infrastructure_node.InfrastructureNode.

        """

        assert type(min_power) is int or type(min_power) is float
        assert type(max_power) is int or type(max_power) is float

        self.id = str(identification)
        self.min_power = min_power
        self.max_power = max_power

        assert parent is None or isinstance(parent, InfrastructureNode)
        self.parent = parent
        if parent is not None:
            parent.add_child(self)
        self.children = []
        self.leafs = None

    def add_child(self, child):
        """Add child to list of all children."""
        assert isinstance(child, InfrastructureNode)
        if child not in self.children:
            self.children.append(child)

    def draw_infrastructure(self):
        """Displays a window with a graph of the infrastructure."""
        graph = nx.Graph()

        # find transformer as the root of the tree
        parent = self
        go_on = True
        while go_on:
            if isinstance(parent.parent, InfrastructureNode):
                parent = parent.parent
                go_on = True
            else:  # type(parent.parent) is None -> root found
                go_on = False
        transformer = parent

        # add transformer to graph
        graph.add_node(transformer)

        # add charging points to graph
        for cs in transformer.children:
            graph.add_node(cs)
            graph.add_edge(transformer, cs)
            # add charging points to graph
            for cp in cs.children:
                graph.add_node(cp)
                graph.add_edge(cs, cp)

        nx.draw(graph, with_labels=True)
        plt.show()

    def get_leaf_nodes(self):
        leafs = []

        def _get_leaf_nodes(node):
            if len(node.children) == 0:
                leafs.append(node)
            for n in node.children:
                _get_leaf_nodes(n)

        _get_leaf_nodes(self)
        return leafs

    def set_up_leafs(self):
        self.leafs = self.get_leaf_nodes()

        def _set_up_leafs(node):
            if len(node.children) == 0:
                return
            for n in node.children:
                n.leafs = n.get_leaf_nodes()
                _set_up_leafs(n)
        _set_up_leafs(self)

    def get_transformer(self):
        parent = self
        go_on = True
        while go_on is True:
            if parent.parent is not None:
                parent = parent.parent
                # If the parent is the Transformer return it
                if isinstance(parent, Transformer):
                    return parent
            else:
                go_on = False

        return None


class Transformer(InfrastructureNode):
    """Represents a transformer. No usability besides having a max and min power.
    Does not have a parent node in infrastructure."""
    counter = 1

    def __init__(self, min_power, max_power):

        identification = 'Transformer_' + str(Transformer.counter)
        Transformer.counter += 1

        super().__init__(identification, min_power, max_power)

    def __str__(self):
        return str(self.id)

    def max_hardware_power(self, power_assigned, preload):
        """Calculate max power assignable to the transformer considering ints limits and already
            assigned power.

            Args:
                power_assigned: (dict): Containing all :obj: `charging_point.ChargingPoint`
                    in the infrastructure and their currently assigned power.
                preload: (float): Preload at the transformer in kWh.

            Returns:
                max_power: (float): Max power that can be assigned to the transformer.
        """
        power_already_assigned = preload
        for leaf in self.leafs:
            if not isinstance(leaf, Storage):
                power_already_assigned += power_assigned[leaf]

        max_power = max(self.max_power - power_already_assigned, 0)
        max_power = floor(max_power * 1000) / 1000

        return max_power


class Storage(InfrastructureNode):
    # ID counter
    counter = 1

    def __init__(self, stationary_battery, transformer):
        assert isinstance(stationary_battery, StationaryBattery)
        self.storage = stationary_battery

        # Power limits
        max_power = self.storage.max_charge_power
        min_power = self.storage.min_charge_power

        # ID
        identification = 'Storage_System ' + str(Storage.counter)
        Storage.counter += 1

        super().__init__(identification, max_power, min_power, parent=transformer)



"""A charging event represents the event of a vehicle charging. It is described in its:
- arrival time: Time the vehicle arrives at the infrastructure.
- parking time: The time the car parks at the infrastructure before it is being driven away.
- """
import datetime


from elvis.vehicle import ElectricVehicle


class ChargingEvent:
    # static field incremented by 1 for every __init__
    counter = 1
    """This class contains relevant parameters to describe a charging event.
    """
    def __init__(self, arrival_time, parking_time, soc, vehicle_type):
        """

        Args:
            arrival_time: (datetime.datetime): Date and time the vehicle arrives at the
                infrastructure.
            parking_time: (float): Number of hours the vehicle is parking.
            soc: (float): SOC of the vehicle at arrival.
            vehicle_type: (:obj: `elvis.vehicle.VehicleType`): Instance of the vehicle describing
                class VehicleType.
        """
        assert isinstance(arrival_time, datetime.datetime)
        assert isinstance(parking_time, (float, int))
        assert isinstance(soc, (float, int)) and (0 <= soc <= 1.0)
        assert isinstance(vehicle_type, ElectricVehicle)

        self.id = 'Charging event: ' + str(ChargingEvent.counter)
        ChargingEvent.counter += 1
        self.arrival_time = arrival_time
        self.parking_time = parking_time
        self.leaving_time = self.arrival_time + datetime.timedelta(hours=self.parking_time)
        self.soc = soc
        self.soc_target = 1.0
        self.vehicle_type = vehicle_type

    def __str__(self):
        print_out = self.id + ', '
        print_out += 'Arrival time: ' + str(self.arrival_time) + ', '
        print_out += 'Parking_time: ' + str(self.parking_time) + ', '
        print_out += 'Leaving_time: ' + str(self.leaving_time) + ', '
        print_out += 'SOC: ' + str(self.soc) + ', '
        print_out += 'SOC target: ' + str(self.soc_target) + ', '
        print_out += 'Connected car: ' + str(self.vehicle_type)
        return print_out

    def to_dict(self, deep=True):
        """Transforms the an instance of ChargingEvent into a dict.

        Args:
            deep: (bool): If True: Instances of vehicle type and battery are converted into dicts
                too. If False: Instances of vehicle type and battery remain objects.
                Use True for storing results, configs, realisations. Use False inside simulation
                so class methods of vehicle type and battery can be used.

        Returns:
            dictionary: (dict): Var names as keys.

        """
        dictionary = self.__dict__.copy()

        dictionary['arrival_time'] = str(self.arrival_time.isoformat())
        dictionary['soc'] = self.soc

        if deep is True:
            dictionary['vehicle_type'] = self.vehicle_type.to_dict()
            del dictionary['leaving_time']
        else:
            dictionary['vehicle_type'] = self.vehicle_type
            dictionary['leaving_time'] = self.leaving_time
        return dictionary

    @staticmethod
    def from_dict(**kwargs):
        """Initialise an instance of ChargingEvent with values stored in a dict.

        Args:

            **kwargs: Arbitrary keyword arguments.
        """

        necessary_keys = ['arrival_time', 'parking_time', 'soc', 'vehicle_type']

        for key in necessary_keys:
            assert key in kwargs, 'Not all necessary keys are included to create a ChargingEvent ' \
                                  'from dict.'

        arrival_time = kwargs['arrival_time']
        parking_time = kwargs['parking_time']
        soc = kwargs['soc']
        vehicle_type = ElectricVehicle.from_dict(**kwargs['vehicle_type'])

        return ChargingEvent(arrival_time, parking_time, soc, vehicle_type)



"""Methods that are elvis related and used in multiple times inside elvis."""

import datetime
import math
import pandas as pd

from elvis.distribution import EquallySpacedInterpolatedDistribution


def create_time_steps(start_date, end_date, resolution):
    """Create list from start, end date and resolution of the simulation period with all individual
    time steps.

    Args:
        start_date: (:obj: `datetime.datetime`): First time stamp.
        end_date: (:obj: `datetime.datetime`): Upper bound for time stamps.
        resolution: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.

    Returns:
        time_steps: (list): Contains time_steps in `datetime.datetime` format

    """
    # Create list containing all time steps as datetime.datetime object
    time_step = start_date
    time_steps = []
    while time_step <= end_date:
        time_steps.append(time_step)
        time_step += resolution

    return time_steps


def num_time_steps(start_date, end_date, resolution):
    """Returns the number of time steps given a period described by its start and end date
    and the resolution.

    Args:
        start_date: (:obj: `datetime.datetime`): First time stamp.
        end_date: (:obj: `datetime.datetime`): Upper bound for time stamps.
        resolution: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.

    Returns:
        num_time_steps: (int): Number of time steps to be simulated.
    """

    return int((end_date - start_date) / resolution) + 1


def transform_data(input_df, resolution, start_date, end_date):
    assert len(input_df) >= 2, "must provide more than two datapoints"
    assert resolution.seconds >= 60, "resolutions lower than one minute not supported"

    # convert to a list of dates and corresponding values
    if isinstance(input_df, pd.DataFrame):
        input_data = []
        for idx, value in input_df.iteritems():
            input_data.append((idx, value))
    else:
        assert isinstance(input_df, list), "input must be list or pandas.DataFrame"
        input_data = input_df

    # get resolution and time frame of input data
    input_start_date = input_data[0][0]
    input_resolution = input_data[1][0] - input_start_date
    input_resolution_seconds = input_resolution.seconds

    # linearly interpolate missing values in input data
    interp_input_data = [input_data[0], input_data[1]]
    for i in range(2, len(input_data)):
        prev = input_data[i - 1]
        curr = input_data[i]
        dist = curr[0] - prev[0]

        # distance between data points is equal to resolution
        if dist == input_resolution:
            interp_input_data.append(curr)
            continue

        # check if distance is evenly divisible by resolution
        rem = dist.seconds % input_resolution_seconds
        if rem != 0:
            raise Exception("inconsistent distance between data points")

        steps = math.floor(dist.seconds / input_resolution_seconds)
        if steps > 0:
            step = (curr[1] - prev[1]) / steps

            for j in range(1, steps):
                interp_input_data.append((prev[0] + (j * input_resolution), prev[1] + (j * step)))

        interp_input_data.append(curr)

    # transform input data to minute resolution
    input_data_minute_res = []

    if input_resolution_seconds > 60:
        # interpolate datapoints to get minute resolution
        interp_steps = math.floor(input_resolution_seconds / 60)
        minute_res = datetime.timedelta(seconds=60)

        for j in range(1, len(interp_input_data)):
            prev = interp_input_data[j - 1][1]
            curr = interp_input_data[j][1]
            step = (curr - prev) / interp_steps

            input_data_minute_res.append(interp_input_data[j - 1])
            for k in range(1, interp_steps):
                input_data_minute_res.append(
                    (interp_input_data[j - 1][0] + (k * minute_res), prev + (k * step)))

    elif input_resolution_seconds < 60:
        # take the average of datapoints within a minute
        curr_sum = interp_input_data[0][1]
        curr_cnt = 1

        curr_date = input_start_date
        curr_start_date = curr_date

        for j in range(1, len(interp_input_data)):
            next_date = interp_input_data[j][0]

            # check if we're at the next minute
            if next_date.minute != curr_date.minute:
                input_data_minute_res.append((curr_start_date, curr_sum / curr_cnt))
                curr_sum = 0
                curr_cnt = 0
                curr_start_date = next_date

            curr_sum += interp_input_data[j][1]
            curr_cnt += 1
            curr_date = next_date

        if curr_sum > 0:
            input_data_minute_res.append((curr_start_date, curr_sum / curr_cnt))
    else:
        input_data_minute_res = interp_input_data

    # now calculate averages for every unique (weekday, month) pair that we have available
    available_datapoints = dict()
    for date, value in input_data_minute_res:
        weekday = date.weekday()
        month = date.month
        key = (weekday, month)

        if key not in available_datapoints:
            available_datapoints[key] = (
                [0 for _ in range(0, 24 * 60)], [0 for _ in range(0, 24 * 60)])

        minute_of_day = date.hour * 60 + date.minute
        available_datapoints[key][0][minute_of_day] += value
        available_datapoints[key][1][minute_of_day] += 1

    for value in available_datapoints.values():
        for j in range(0, len(value[0])):
            if value[1][j] <= 1:
                continue

            value[0][j] /= value[1][j]

    # uncomment to export data for all (weekday, month) pairs
    # for key, value in available_datapoints.items():
    #    export_csv(value[0], "./input_data_averaged/" + str(key[0]) + "_" + str(key[1]) + ".csv")

    # build the result by finding the closest (weekday, month) pair for every required day
    result_data = []

    curr_date = start_date
    curr_day_data = None
    prev_weekday = None

    while curr_date < end_date:
        curr_weekday = curr_date.weekday()
        if prev_weekday is None or curr_weekday != prev_weekday:
            curr_month = curr_date.month
            prev_weekday = curr_weekday

            if (curr_weekday, curr_month) in available_datapoints:
                # use the available data for the day
                curr_day_data = available_datapoints[(curr_weekday, curr_month)]
            else:
                # find the closest available data point
                min_dist = math.inf
                for key, value in available_datapoints.items():
                    dist = abs(key[0] - curr_weekday) + abs(key[1] - curr_month)
                    if dist >= min_dist:
                        continue

                    min_dist = dist
                    curr_day_data = value

        minute_of_day = curr_date.hour * 60 + curr_date.minute
        result_data.append((curr_date, curr_day_data[0][minute_of_day]))

        curr_date += resolution

    return result_data


def adjust_resolution(preload, res_data, res_simulation):
    """Adjusts res_data of the transformer preload to the simulation res_data.

    Args:
        preload: (list): Containing the transformer preload in "wrong"
            res_data.
        res_data: (datetime.timedelta): Time in between two adjacent data points of
            transformer preload with "wrong" res_data.
        res_simulation: (datetime.timedelta): Time in between two adjacent time steps in
            the simulation.

    Returns:
        transformer_preload_new_res: (list): Transformer preload with linearly interpolated data
            points having the res_data of the simulation.
        """

    x_values = list(range(len(preload)))
    distribution = EquallySpacedInterpolatedDistribution.linear(
        list(zip(x_values, preload)), None)

    coefficient = res_simulation / res_data
    x_values_new_res = list(range(math.ceil(len(preload) * 1 / coefficient)))
    x_values_new_res = [x * coefficient for x in x_values_new_res]

    transformer_preload_new_res = []
    for x in x_values_new_res:
        transformer_preload_new_res.append(distribution[x])

    return transformer_preload_new_res


def repeat_data(preload, num_simulation_steps):
    """Repeats the transformer preload data until there are as many values as there are
    simulation steps.

    Args:
        preload: (list): Containing the data (floats) to be repeated.
        num_simulation_steps: (int): Number of simulation steps and expected length of
            the transformer preload after it is repeated.

    Returns:
        transformer_preload_repeated: (list): Repeated values. len() = num_simulation_steps.
        """

    n = math.floor(num_simulation_steps / len(preload))

    transformer_preload_repeated = preload * n

    values_to_add = num_simulation_steps - len(transformer_preload_repeated)

    transformer_preload_repeated += preload[:values_to_add]

    return transformer_preload_repeated


def floor(value, decimals=3):
    """Floors a value to 3 decimals."""

    assert isinstance(decimals, int), 'Decimals must be of type int'

    coeff = 10**decimals
    value = math.floor(value * coeff) / coeff

    return value



"""Representation of a charging station of the infrastructure.

A charging station is the point of coupling between charging point (connection to the car)
and the grid. In LV grids the charging station immediately connects to the grid in MV grids a
transformer is needed in between.
- The charging station may consist of multiple charging points.
TODO:
    - It may be automated: After a car is charged it automatically connects itself to another car if
    available.
    - It may be inductive: No physical connection between car and charging point is needed:
    lower efficiency.
- Or manual: User have to connect and disconnect their vehicle by hand: Longer phases of inactivity.
TODO:
    - It can be operated with direct current or alternating current where different efficiencies
    have to be considered.
    - Its power is the sum over the power of its charging points.
    - It is limited by its power limits where the lower limit has to be understood as at least the
    lower limit or zero.
"""

from elvis.utility.elvis_general import floor
from elvis.infrastructure_node import InfrastructureNode


class ChargingStation(InfrastructureNode):
    """Models a charging point of the charging ingrastructure.
    """

    counter = 1

    def __init__(self, min_power, max_power, parent):
        # id
        identification = 'cs' + str(ChargingStation.counter)
        ChargingStation.counter += 1

        # power limits
        super().__init__(identification, min_power, max_power, parent=parent)

        # working principle
        self.manual = True
        self.automated = False
        self.inductive = False

        # power type AC = True, DC = False
        self.ac_power_supply = True

        self.efficiency = 1

    def __str__(self):
        return str(self.id)

    def max_hardware_power(self, power_assigned):
        """Determine max power possible to be assigned dependent on charging station limits and
            power already assigned to charging points connected to the charging station.
            Current simplification: Charging points always are directly connected to charging
            stations.

            Args:
                power_assigned: (dict): Contains all :obj: `charging_point.ChargingPoint` and
                    their currently already assigned power.
        """

        power_cps = 0
        for leaf in self.leafs:
            power_cps += power_assigned[leaf]

        max_power = self.max_power - power_cps
        max_power = floor(max_power)

        return max_power



"""Create infrastructure: connect charging points to transformer and charging points to
    charging stations."""

from copy import deepcopy

from elvis.charging_station import ChargingStation
from elvis.charging_point import ChargingPoint
from elvis.infrastructure_node import Transformer, Storage
from elvis.battery import StationaryBattery


def set_up_infrastructure(infrastructure):
    """Reads in infrastructure layout as a dict and converts it into a tree shaped infrastructure
        design with nodes representing the hardware components:
            transformer: :obj: `infrastructure_node.Transformer`
            charging station: :obj: `charging_station.ChargingStation`
            charging point: :obj: `charging_point.ChargingPoint`
        Returns all charging points.

    Args:
        infrastructure: (dict): Contains more nested dicts with the values to initialise the
            infrastructure nodes.
            Dict design:
                transformer: {min_power(float), max_power(float), infrastructure(list)}

            For each instance of charging station in infrastructure:
                charging sation: {min_power(float), max_power(float), charging_points(list)}

            For each instance of charging_point in charging_stations:
                charging_point: {min_power(float), max_power(float)}

    Returns:
        charging_points: (list): Contains n instances of :obj: `charging_point.ChargingPoint`.
    """
    charging_points = []
    transformer = None
    # build infrastructure and create node instances
    # Add transformer
    for __transformer in infrastructure['transformers']:
        transformer = Transformer(__transformer['min_power'], __transformer['max_power'])
        # Add all charging points and their charging points
        for __charging_station in __transformer['charging_stations']:
            charging_station = ChargingStation(__charging_station['min_power'],
                                               __charging_station['max_power'],
                                               transformer)

            # Add all charging points of current charging point
            for __charging_point in __charging_station['charging_points']:
                __charging_point = ChargingPoint(__charging_point['min_power'],
                                                 __charging_point['max_power'],
                                                 charging_station)

                charging_points.append(__charging_point)
        if 'storage' in __transformer.keys():
            __battery = __transformer['storage']
            if 'efficiency' in __battery.keys():
                bat_eff = __battery['efficiency']
            else:
                bat_eff = 1
            battery = StationaryBattery(capacity=__battery['capacity'],
                                        max_charge_power=__battery['max_power'],
                                        min_charge_power=__battery['min_power'],
                                        efficiency=bat_eff)
            storage = Storage(battery, transformer)

    assert isinstance(transformer, Transformer), 'Invalid infrastructure dict.'
    transformer.set_up_leafs()
    # transformer.draw_infrastructure()

    return charging_points


def wallbox_infrastructure(num_cp, power_cp, num_cp_per_cs=1, power_cs=None,
                           power_transformer=None, min_power_cp=0, min_power_cs=0,
                           min_power_transformer=0):
    """Builds an Elvis conform infrastructure.

    Args:
        num_cp: (int): Number of charging points.
        num_cp_per_cs: (int) Number of charging points per charging station.
        power_cp: (int or float): Max power per charging point.
        power_cs: (int or float): Max power of the charging station.
        power_transformer: (int or float): Max power of the transformer.
        min_power_cp: (int or float): Minimum power (if not 0) for the charging point.
        min_power_cs: (int or float): Minimum power (if not 0) for the charging station.
        min_power_transformer: (int or float) : Minimum power (if not 0) for the charging station.

    Returns:
        infrastructure: (dict)
    """
    # Validate Input
    assert isinstance(num_cp_per_cs, int), 'Only integers are allowed for the number of ' \
                                           'charging points per charging station.'
    num_cs = int(num_cp / num_cp_per_cs)
    assert num_cp % num_cp_per_cs == 0, 'Only integers are allowed for the number of charging ' \
                                        'stations. The passed number of charging points and the ' \
                                        'number of charging points per charging station are not ' \
                                        'realisable.'
    num_cs = int(num_cp / num_cp_per_cs)

    msg_power_numeric = 'The power assigned to the infrastructure elements must be of type int or' \
                        'float. The power must be bigger than 0.'

    if power_cs is None:
        power_cs = power_cp * num_cp_per_cs
    assert isinstance(power_cp, (int, float)), msg_power_numeric
    assert isinstance(power_cs, (int, float)), msg_power_numeric
    if power_transformer is None:
        power_transformer = num_cs * num_cp_per_cs * power_cp
    assert isinstance(power_transformer, (int, float)), msg_power_numeric
    assert isinstance(min_power_cp, (int, float)), msg_power_numeric
    assert isinstance(min_power_transformer, (int, float)), msg_power_numeric
    assert isinstance(min_power_cs, (int, float)), msg_power_numeric

    msg_min_max = 'The min power assigned to a infrastructure node must be lower than the ' \
                  'max power.'
    assert power_cp >= min_power_cp, msg_min_max
    assert power_cs >= min_power_cs, msg_min_max
    assert power_transformer >= min_power_transformer, msg_min_max

    transformer = {'charging_stations': [], 'id': 'transformer1',
                   'min_power': min_power_transformer, 'max_power': power_transformer}

    charging_point = {'min_power': min_power_cp, 'max_power': power_cp}

    charging_station = {'min_power': min_power_cs, 'max_power': power_cs,
                        'charging_points': []}

    for i in range(num_cs):
        charging_station_temp = deepcopy(charging_station)
        charging_station_temp['id'] = 'cs' + str(i + 1)
        for j in range(num_cp_per_cs):
            charging_point_temp = deepcopy(charging_point)
            charging_point_temp['id'] = 'cp' + str(i * num_cp_per_cs + j+1)
            charging_station_temp['charging_points'].append(charging_point_temp)
        transformer['charging_stations'].append(charging_station_temp)

    infrastructure = {'transformers': [transformer]}

    return infrastructure




import math


class Distribution:
    """Represents a distribution of some x value to a y value."""

    def __getitem__(self, key):
        raise NotImplementedError()

    @property
    def bounds(self):
        raise NotImplementedError()

    @property
    def min_x(self):
        return bounds["x"]["min"]

    @property
    def max_x(self):
        return bounds["x"]["max"]

    @property
    def min_y(self):
        return bounds["y"]["min"]

    @property
    def max_y(self):
        return bounds["y"]["max"]


class NormalDistribution(Distribution):
    """A normal distribution."""

    def __init__(self, mu, sigma):
        self.mu = mu
        self.sigma = sigma

        # 1 / (sigma*sqrt(2*pi))
        self.fac = 1.0 / (sigma * math.sqrt(2.0 * math.pi)) 

    def __getitem__(self, key):
        # ((x-mu)/sigma)^2
        exp = math.pow((key - self.mu) / self.sigma, 2.0)

        # fac * e^(-.5*exp)
        return self.fac * math.exp(-0.5 * exp)

    @property
    def bounds(self):
        return {
            "x": {
                "min": -math.inf,
                "max": math.inf
            },
            "y": {
                "min": 0,
                "max": 1,
            }
        }


class InterpolatedDistribution(Distribution):
    """A distribution that generates new values using some form of interpolation of a set of given points."""

    def __init__(self, points, bounds, interpolate):
        self.points = points
        self._bounds = bounds
        self.interpolate = interpolate

    @staticmethod
    def linear(points, bounds):
        return InterpolatedDistribution(points, bounds, InterpolatedDistribution._linear_interpolation)

    @staticmethod
    def _linear_interpolation(y0, y1, offset):
        return y0 + (y1 - y0) * offset

    def __getitem__(self, key):
        if key < self.points[0][0]:
            return self.points[0][1]

        if key >= self.points[-1][0]:
            return self.points[-1][1]

        # find the closest point we have.
        min_i = 0

        for i, pt in enumerate(self.points):
            if pt[0] >= key:
                min_i = i
                break

        p0 = self.points[min_i - 1]
        p1 = self.points[min_i]
        offset = (key - p0[0]) / (p1[0] - p0[0])

        return self.interpolate(p0[1], p1[1], offset)

    @property
    def bounds(self):
        return self._bounds


class EquallySpacedInterpolatedDistribution(Distribution):
    """A distribution that generates new values using some form of interpolation of a set of given points,
     whose x values are the same distance apart."""

    def __init__(self, points, bounds, interpolate):
        self.points = points
        self._bounds = bounds
        self.interpolate = interpolate
        self.distance_between_points = abs(points[1][0] - points[0][0])

    @staticmethod
    def linear(points, bounds):
        return EquallySpacedInterpolatedDistribution(points, bounds, InterpolatedDistribution._linear_interpolation)

    @staticmethod
    def _linear_interpolation(y0, y1, offset):
        return y0 + (y1 - y0) * offset

    def __getitem__(self, key):
        if key < self.points[0][0]:
            return self.points[0][1]

        if key >= self.points[-1][0]:
            return self.points[-1][1]

        # find the closest point we have.
        min_i = math.ceil((key - self.points[0][0]) / self.distance_between_points)

        p0 = self.points[min_i - 1]
        p1 = self.points[min_i]
        offset = (key - p0[0]) / (p1[0] - p0[0])

        return self.interpolate(p0[1], p1[1], offset)

    @property
    def bounds(self):
        return self._bounds



"""Used to generate charging events based on distributions or datapoints
of measured charging events.

TODO:
    - Ensure other than hourly distributions work
    - Add parking time, car_type, arrival_soc, target_soc
    - Enable conversion of measured charging events to
    synthetic charging events
    - Catch and prevent errors, add raised to docstrings
"""
import math
import datetime
import numpy as np

import elvis.distribution as distribution
import elvis.charging_event as charging_event

from elvis.utility.walker import WalkerRandomSampling
from sklearn.mixture import GaussianMixture


def time_stamp_to_hours(time_stamps):
    """Calculates for each time stamp in a list the amount of hours passed
        since the beginning of the hour of the first time stamp.
    Args:
        time_stamps (list): Containing the time stamps as :obj: `datetime.datetime`.

    Returns:
        list: Hours passed.
    """
    start = time_stamps[0]
    hours_passed = []
    for time_stamp in time_stamps:
        delta = time_stamp - start
        hour = delta.days * 24 + delta.seconds / 3600 + delta.microseconds / 3600 / 1000 / 1000
        # add offset
        hour += start.minute / 60 + start.second / 3600 + start.microsecond / 3600 / 1000 / 1000
        hours_passed.append(hour)
    return hours_passed


def hours_to_time_stamps(hours, start):
    """Converts hour stamps to time stamps. Hour stamps measuring the time
        passed sinde the beginning of the first time stamp.

    Args:
          hours (list): Containing the hours to convert.
          start (:obj: `datetime.datetime`): First time stamp.

    Returns:
        list: Time stamps.
    """
    # Define the corresponding time stamp to hour = 0.
    # Hour = 0 is the beginning of the hour of the first time stamp.
    hour0_corr = datetime.datetime(start.year, start.month, start.day, start.hour)

    time_stamps = []
    for hour in hours:
        time_stamps.append(hour0_corr + datetime.timedelta(hours=hour))

    return time_stamps


def align_distribution(distr, first_time_stamp, last_time_stamp):
    """Receives weekly distribution in hourly resolution besides a starting and an
    ending time_stamp. Shifts and multiplies distribution to allign it to the
     period of the time_stamps.

    Args:
        distr (list): Containing hourly probabilities.
        first_time_stamp (:obj: `datetime.datetime`): Start of period.
        last_time_stamp (:obj: `datetime.datetime`): End of period.

    Returns:
        aligned_distribution: (list): Containing the probabilities alligned to period.
        difference: Difference between first used 'time stamp' of the distribution and first
            time stamp of the simulation in hours.
    """
    hours = first_time_stamp.weekday() * 24 + first_time_stamp.hour
    minutes = first_time_stamp.minute
    seconds = first_time_stamp.second
    microseconds = first_time_stamp.microsecond

    offset = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds,
                                microseconds=microseconds)

    seconds_per_week = 7 * 24 * 3600
    num_values = len(distr)
    seconds_per_value = seconds_per_week/num_values

    starting_pos = math.floor(offset.total_seconds()/seconds_per_value)
    difference = (offset.total_seconds()/seconds_per_value - starting_pos) / 3600

    period = last_time_stamp - first_time_stamp
    # Upward estimate of the needed length of the distribution
    total_weeks = math.ceil(period.total_seconds() / seconds_per_week)

    aligned_distribution = distr[starting_pos:] + distr * total_weeks
    return aligned_distribution, difference


def create_vehicle_arrivals(arrival_distribution, num_charging_events, time_steps):
    """Creates vehicle arrival times.

    Args:
        arrival_distribution (list): Containing hourly arrival probabilities for one week.
        num_charging_events: (int): Number of charging events per week.
        time_steps (list): List containing all time steps as :obj: `datetime.datetime` object.

    Returns:
        list: Arrival times.

    ToDo:
        seeding seems to not work properly. With fixed seed small changes are still there,
        changing the seed has a huge impact though.
    """

    coefficient = 168 / len(arrival_distribution)

    # Rearrange arrival distribution so it starts with first hour of simulation time
    arrival_distribution, difference = align_distribution(arrival_distribution, time_steps[0],
                                                          time_steps[-1])

    # generate x-values (hours away from first time step) of the distribution
    hour_stamps = [x * coefficient - difference for x in range(len(arrival_distribution))]
    # Create distribution based on reordered arrival distribution
    dist = distribution.EquallySpacedInterpolatedDistribution.linear(
        list(zip(hour_stamps, arrival_distribution)), None)

    # Calculate position of each time step at arrival distribution
    corr_position = time_stamp_to_hours(time_steps)

    # Get arrival probablity for each time step of the simulation
    arrival_probability = []
    for pos in corr_position:
        arrival_probability.append(dist[pos])
    # Normalize probability
    cumsum = sum(arrival_probability)
    arrival_probability = [x / cumsum for x in arrival_probability]

    # Get on average num_charging_events arrivals per week
    period = time_steps[-1] - time_steps[0]
    num_weeks = period.total_seconds() / 7 / 24 / 3600
    corr_times = np.random.choice(corr_position, p=arrival_probability,
                                  size=math.ceil(num_charging_events * num_weeks))

    # Convert hours back to time_stamps
    arrivals = hours_to_time_stamps(corr_times, time_steps[0])

    return sorted(arrivals)


def create_charging_events_from_weekly_distribution(
        arrival_distribution, time_steps, num_charging_events, mean_park, std_deviation_park,
        mean_soc, std_deviation_soc, vehicle_types, max_parking_time):

    """Create all charging events for the simulation period.

    Args:
        arrival_distribution: (list): Containing hourly data for the arrival probabilities for one
        week.
        time_steps: (list): Contains time_steps in `datetime.datetime` format
        num_charging_events: (int): Number of charging events to be generated.
        mean_park: (float): Mean of the gaussian distribution the parking time is generated from.
        std_deviation_park: (float): Standard deviation of the gaussian distribution the parking
            time is generated from.
        mean_park: (float): Mean of the gaussian distribution the SOC is generated from.
        std_deviation_park: (float): Standard deviation of the gaussian distribution the SOC
            is generated from.
        mean_soc: (float): Limits: [0, 1]. State of charge of the cars when arriving on average.
        std_deviation_soc: (float): Standard deviation of the arrival SOC.
        vehicle_types: (list): Containing all instances of :obj: `elvis.vehicle.ElectricVehicle`
        max_parking_time: (float): Maximum time a car is allowed to stay at the charging
            infrastructure in hours.

    Returns:
        (list): containing num_charging_events instances of `ChargingEvent`.

    """
    msg_no_vehicles = 'At least one vehicle type is necessary to generate charging events.'
    assert len(vehicle_types) > 0, msg_no_vehicles

    arrivals = create_vehicle_arrivals(arrival_distribution, num_charging_events, time_steps)
    weights = [vehicle_type.probability for vehicle_type in vehicle_types]

    walker = WalkerRandomSampling(weights, keys=vehicle_types)

    charging_events = []
    parking_time_samples = np.random.normal(mean_park, std_deviation_park, len(arrivals)).tolist()
    soc_samples = np.random.normal(mean_soc, std_deviation_soc, len(arrivals)).tolist()
    vehicle_type_samples = walker.random(count=len(arrivals)).tolist()
    for arrival in arrivals:
        parking_time = min(max(0, parking_time_samples.pop()), max_parking_time)
        soc = min(max(0, soc_samples.pop()), 1)
        vehicle_type = vehicle_type_samples.pop()
        charging_events.append(charging_event.ChargingEvent(arrival, parking_time, soc,
                                                            vehicle_type))

    return charging_events


def init_gmm(means, weights, covariances):
    """
        Initialise the GMM.

    Args:
        means: (list): Mean value for each mixture component.
        weights: (list): Mixing weights for each mixture component.
        covariances: (list): Covariance for mixture component. In the used scikit-learn model
            a covariance_type of "full" is used.

    Returns:
        gmm: (:obj:): Scikit-learn gaussian mixture model.
    """

    assert 0.99 < sum(weights) < 1.01, 'GMM weights must add up to 1 (1 % tolerance)'
    assert len(weights) == len(means) == len(covariances), 'Parameters for GMM must be of same ' \
                                                           'length'
    assert all(0 < x[0] < 168 for x in means), 'The first dimension of the means is expected to ' \
                                               'be the arrival time in hours counting from ' \
                                               'Monday 0:00. Values should therefore be >0 and <168'

    # 2 dimensions: arrival time, parking time -> 2x2 covariance matrices
    for i in covariances:
        for j in i:
            assert len(i) == len(j) == 2, 'GMM covariances must be quadratic with 2 rows and 2 cols'

    # numpy arrays of parameters
    means_np = np.asarray(means)
    weights_np = np.asarray(weights)
    covariances_np = np.asarray(covariances)

    # scikit-learn GMM
    gmm = GaussianMixture()

    # initialise parameters
    gmm.means_ = means_np
    gmm.weights_ = weights_np
    gmm.covariances_ = covariances_np

    return gmm


def weeks_to_sample(time_steps):
    """
    Calculates the weeks to sample for a GMM model based on the time steps of the simulation.

    Args:
        time_steps: (list): Contains time_steps in `datetime.datetime` format

    Returns:
        num_weeks: (int): # weeks to sample
    """
    # Total duration of simulation period in hours
    last = time_steps[-1]
    first = time_steps[0]
    total_hours = (last - first).total_seconds() / 3600

    num_weeks = math.ceil(total_hours / 168)

    # If the first time stamp is later in the week add one week to sample
    if first.weekday() > last.weekday():
        num_weeks += 1
    elif first.weekday() == last.weekday():
        if first.time() > last.time():
            num_weeks += 1

    return num_weeks


def create_charging_events_from_gmm(time_steps, num_charging_events, means, weights,
                                    covariances, vehicle_types, max_parking_time, mean_soc,
                                    std_deviation_soc):
    """

    Args:
        time_steps: (list): Contains time_steps in `datetime.datetime` format
        num_charging_events: (int): Number of charging events to be generated.
        means: (list): Mean value for each mixture component.
        weights: (list): Mixing weights for each mixture component.
        covariances: (list): Covariance for mixture component. In the used scikit-learn model
            a covariance_type of "full" is used.
        vehicle_types:  (list): Containing all instances of :obj: `elvis.vehicle.ElectricVehicle`
        max_parking_time: (float): Maximum time a car is allowed to stay at the charging
            infrastructure in hours.
        mean_soc: (float): Limits: [0, 1]. State of charge of the cars when arriving on average.
        std_deviation_soc: (float): Standard deviation of the arrival SOC.

    Returns:
        (list): containing num_charging_events instances of `ChargingEvent`.
    """

    gmm = init_gmm(means, weights, covariances)

    num_weeks = weeks_to_sample(time_steps)

    week_offset = 0
    samples = []
    rounding_const = 1/((time_steps[1] - time_steps[0]).total_seconds() / 3600)
    while num_weeks > week_offset / 168:
        temp = gmm.sample(num_charging_events)
        # Discard cluster assignments
        temp = temp[0]
        temp = temp.tolist()
        temp.sort()
        for sample in temp:
            sample[0] += week_offset
            # to ensure arrivals fall on time stamps
            sample[0] = math.ceil(sample[0] * rounding_const) / rounding_const
            samples.append(sample)
        week_offset += 168

    # hours from monday of the first week until the beginning of the simulation
    day_offset = time_steps[0].weekday() * 24
    first_step_hours = day_offset + time_steps[0].hour + time_steps[0].minute / 60 + \
                       time_steps[0].second / 60 / 60
    # Remove samples before first simulation step
    # Ensure stop is < first_step_hours for initialisation
    stop = first_step_hours - 1
    i = 0
    while stop < first_step_hours:
        stop = samples[i][0]
        i += 1
    # Discard all samples before first time stamp
    samples = samples[i-1:]

    # simulation duration in hours
    sim_dur_hours = (time_steps[-1] - time_steps[0]).total_seconds() / 3600
    i = 0
    # ensure initial stop > sim_dur_hours
    stop = sim_dur_hours + 1
    while stop >= sim_dur_hours:
        i += 1
        stop = samples[-(i-1)][0]
    # discard all charging events after simulation period
    if i > 1:
        samples = samples[:-(i-1)]

    # transform arrival time in hours into datetime.datetime object
    # initialise reference point Monday 0:00 of first simulation week
    ref_date = time_steps[0] - datetime.timedelta(hours=first_step_hours)
    for sample in samples:
        sample_hour = sample[0]
        sample[0] = ref_date + datetime.timedelta(hours=sample_hour)
        # ensure 1 min < parking time < max_parking_time
        sample[1] = max(min(sample[1], max_parking_time), 0.0167)

    walker_weights = [vehicle_type.probability for vehicle_type in vehicle_types]
    walker = WalkerRandomSampling(walker_weights, keys=vehicle_types)

    charging_events = []
    soc_samples = np.random.normal(mean_soc, std_deviation_soc, len(samples)).tolist()
    vehicle_type_samples = walker.random(count=len(samples)).tolist()
    for sample in samples:
        soc = min(max(0, soc_samples.pop()), 1)
        vehicle_type = vehicle_type_samples.pop()
        charging_events.append(charging_event.ChargingEvent(sample[0], sample[1], soc,
                                                            vehicle_type))

    return charging_events



"""Simulation of the behavior of the charging infrastructure.

- TODO: What if connection points/charging points are not type manual but type automated."""

import datetime
import logging

from elvis.utility.elvis_general import create_time_steps
from elvis.set_up_infrastructure import set_up_infrastructure
from elvis.sched.schedulers import Uncontrolled, FCFS
from elvis.waiting_queue import WaitingQueue
from elvis.result import ElvisResult
from elvis.config import ScenarioRealisation, ScenarioConfig
from elvis.infrastructure_node import Storage


def handle_car_arrival(free_cps, busy_cps, event, waiting_queue, counter_rejections,
                       within_opening_hours, log):
    """Connects car to charging point, to the queue or send it off.

    Args:
        free_cps: (set): Containing all :obj: `charging_points.ChargingPoint`
            of the infrastructure that are currently available.
        busy_cps: (set): Containing all :obj: `charging_points.ChargingPoint`
            of the infrastructure that are currently busy.
        event: (:obj: `charging_event.ChargingEvent`): Arriving charging event.
        waiting_queue: (:obj: `queue.WaitingQueue`): Containing the waiting vehicles
        counter_rejections: (int): Counts how many cars can not connect to a charging point nor to
            the waiting_queue.
        within_opening_hours: (bool): True if currently within opening hours.
        log: (bool): True if actions taken shall be logged.

    """

    # connect the arrival to an available charging point if possible
    if len(free_cps) > 0 and within_opening_hours:
        # Get random free charging point and remove it from set
        cp = free_cps.pop()
        cp.connect_vehicle(event)
        if log:
            logging.info(' Connect: %s', cp)
        # Put charging point to busy set
        busy_cps.add(cp)
        return waiting_queue, counter_rejections

    # if no free charging point available put arrival to queue
    # if not full and queue is considered in simulation (maxsize > 0)
    if not waiting_queue.size() == waiting_queue.maxsize and waiting_queue.maxsize > 0 and \
            within_opening_hours:
        waiting_queue.enqueue(event)
        if log:
            logging.info(' Put %s to queue.', event)
    else:
        if log and within_opening_hours:
            logging.info(' WaitingQueue is full -> Reject: %s', event)
        elif log and not within_opening_hours:
            logging.info('Out of opening hours.')
        counter_rejections += 1
    return waiting_queue, counter_rejections


def update_queue(waiting_queue, current_time_step, by_time, within_opening_hours, log):
    """Removes cars that have spent their total parking time in the waiting queue and
        therefore must leave.

    Args:
        waiting_queue: (:obj: `queue.WaitingQueue`): Charging events waiting for available
        charging point.
        current_time_step: (:obj: `datetime.datetime`): current time step of the simulation.
        by_time: (bool): True if cars shall be disconnected with respect to their parking time.
        False if cars shall be disconnected due to their SOC target.
        within_opening_hours: (bool): True if currently within opening hours.
        log: (bool): True if actions taken shall be logged.

    Returns:
        updated_queue: (:obj: `queue.WaitingQueue`): WaitingQueue without removed cars
            that are overdue.
    """
    if not within_opening_hours:
        waiting_queue.empty()
        logging.info(' Remove all vehicles from queue -> out of opening hours.')
    if by_time is True and current_time_step >= waiting_queue.next_leave:
        to_delete = []
        for i in range(waiting_queue.size()):
            event = waiting_queue.queue[i]

            # If waiting time is not longer than parking time remain in queue
            if event.leaving_time <= current_time_step:
                to_delete.insert(0, i)

        if len(to_delete) > 0:
            if log:
                for i in to_delete:
                    logging.info(' Remove: %s from queue.', waiting_queue.queue.pop(i))

        waiting_queue.determine_next_leaving_time()


def update_cps(free_cps, busy_cps,
               waiting_queue, current_time_step, by_time, within_opening_hours, log):
    """Removes cars due to their parking time or their SOC limit and updates the charging points
    respectively.

    Args:
        free_cps: (set): Containing all :obj: `charging_point.ChargingPoint`
            of the infrastructure that are currently available.
        busy_cps: (set): Containing all :obj: `charging_point.ChargingPoint`
            of the infrastructure that are currently busy.
        waiting_queue: (:obj: `queue.WaitingQueue`): waiting vehicles/charging events.
        current_time_step: (:obj: `datetime.datetime`): current time step.
        by_time: (bool): Configuration class instance.
        within_opening_hours: (bool): True if currently within opening hours.
        log: (bool): True if actions taken shall be logged.
    """
    # if outside of opening hours disconnect all vehicles
    if not within_opening_hours:
        cps_to_remove = []
        for cp in busy_cps:
            cp.disconnect_vehicle()
            cps_to_remove.append(cp)
            if log:
                logging.info(' Disconnect: %s', cp)
        for cp in cps_to_remove:
            busy_cps.remove(cp)
            free_cps.add(cp)
        return

    temp_switch_cps = []
    # if parking time is overdue: disconnect vehicle
    if by_time is True:
        for cp in busy_cps:
            connected_vehicle = cp.connected_vehicle

            if connected_vehicle['leaving_time'] <= current_time_step:
                if log:
                    logging.info(' Disconnect: %s', cp)
                cp.disconnect_vehicle()
                temp_switch_cps.append(cp)

                # immediately connect next waiting car
                if waiting_queue.size() > 0:
                    cp.connect_vehicle(waiting_queue.dequeue())
                    # temporary store cps to switch later so busy_cp set does not change size
                    temp_switch_cps = temp_switch_cps[:-1]
                    # Put charging point from available to busy
                    if log:
                        logging.info(' Connect: %s from queue.', cp)

        for cp in temp_switch_cps:
            busy_cps.remove(cp)
            free_cps.add(cp)

    # if SOC limit is reached: disconnect vehicle
    else:
        for cp in busy_cps:
            connected_vehicle = cp.connected_vehicle

            soc = connected_vehicle['soc']
            soc_target = connected_vehicle['soc_target']

            if round(soc, 3) >= soc_target:
                if log:
                    logging.info(' Disconnect: %s', cp)
                cp.disconnect_vehicle()

                temp_switch_cps.append(cp)

                # immediately connect next waiting car
                if waiting_queue.size() > 0:
                    cp.connect_vehicle(waiting_queue.dequeue())
                    # temporary store cps to switch later so set does not change size
                    temp_switch_cps = temp_switch_cps[:-1]
                    if log:
                        logging.info(' Connect: %s from queue.', cp)
                # Put charging point from available to busy
        for cp in temp_switch_cps:
            busy_cps.remove(cp)
            free_cps.add(cp)


def charge_connected_vehicles(assign_power_cps, busy_cps, res, log):
    """Change SOC of connected vehicles based on power assigned by scheduling policy.

    Args:
        assign_power_cps: (dict): keys=charging points, values=power to be assigned.
        busy_cps: list of all charging points that currently have a connected
            vehicle.
        res: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.
        log: (bool): Flag denoting whether a logging entry is supposed to take place.

    Returns: None
    """

    for cp in busy_cps:
        power = assign_power_cps[cp]
        vehicle = cp.connected_vehicle
        soc_before = vehicle['soc']
        if vehicle is None:
            raise TypeError

        cp.charge_vehicle(power, res)

        if log:
            logging.info('At charging point %s the vehicle SOC has been charged from %s to %s. '
                         'The power assigned is: %s', cp, soc_before, vehicle['soc'],
                         str(power))


def charge_storage(assign_power, preload, step_length):
    """Charges/discharges the storage and returns the realised (dis)charging power.

    Args:
        assign_power: (dict): power assigned to storage and cps
        preload: (float): Preload at current time step.
        step_length: (:obj: `datetime.timedelta`): Resolution of the simulation denoting the
                time in between two adjacent time steps.
    Returns:
        assign_power: (dict): keys=storage, value=realised power

    """
    assign_power_storage = assign_power['storage']
    assign_power_cps = assign_power['cps']
    for storage_system in assign_power_storage:
        power_to_storage = assign_power_storage[storage_system]
        storage = storage_system.storage

        # Storage is not needed to charge cps so see if charging is possible
        if power_to_storage == 0:
            transformer = storage_system.get_transformer()
            power_available = transformer.max_hardware_power(assign_power_cps, preload)
            # Charge storage depending on available power and its limits
            power_charged = storage.charge(power_available, step_length)
            # Update assigned power
            assign_power['storage'][storage_system] = power_charged
        # Try to discharge with assigned power
        else:
            storage.discharge(abs(power_to_storage), step_length)

    return assign_power


def update_last_charged(charging_times, assign_power_cps, time_step):
    """
    Updates the dict containing parking events with their start time and the last time they were
    charged based on the currently assigned power.

    Args:
        charging_times: (dict): Containing each car as keys and their connection time and the last
            time they were charged.
        assign_power_cps: (dict): Containing the powers assigned to each busy charging point at the
            current time step.
        time_step: (:obj: `datetime.datetime`): Current time step.

    Returns:
        charging_times_updated: (dict): Added cars that have not been charged before and updates
            last charged if car has been charged again.
    """

    cps_with_power = {k: v for k, v in assign_power_cps.items() if v > 0}

    charging_times_updated = charging_times
    for cp in cps_with_power:
        ce = cp.connected_vehicle['id']
        if ce not in charging_times_updated:
            charging_times_updated[ce] = {'arrival': time_step, 'last_charged': time_step}
        else:
            charging_times_updated[ce]['last_charged'] = time_step

    return charging_times_updated


def simulate(scenario, start_date=None, end_date=None, resolution=None, realisation_file_name=None,
             log = False):
    """Main simulation loop.
    Iterates over simulation period and simulates the infrastructure.

    Args:
        scenario: (:obj: `elvis.scenario.ScenarioRealisation`): Scenario to be simulated.
        start_date: (:obj: `datetime.datetime`): First time stamp.
        end_date: (:obj: `datetime.datetime`): Upper bound for time stamps.
        resolution: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.

    Returns:
        result: (:obj: `elvis.result.ElvisResult`): Contains the results of the simulation.
    """
    # set up charging points in result to store assigned powers
    if realisation_file_name is None:
        results = ElvisResult()
    else:
        results = ElvisResult(scenario, realisation_file_name)

    for progress in simulate_async(scenario, results, start_date, end_date, resolution, log):
        print('Progress: ' + str(round(progress*100, 0)) + ' %')

    return results


def simulate_async(scenario, results, start_date=None, end_date=None, resolution=None, log = False):
    """Main simulation loop.
    Iterates over simulation period and simulates the infrastructure.

    Args:
        scenario: (:obj: `elvis.scenario.ScenarioRealisation`): Scenario to be simulated.
        start_date: (:obj: `datetime.datetime`): First time stamp.
        end_date: (:obj: `datetime.datetime`): Upper bound for time stamps.
        resolution: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.

    Returns:
        result: (:obj: `elvis.result.ElvisResult`): Contains the results of the simulation.
    """
    # ---------------------Initialisation---------------------------
    # if input is instance of ScenarioConfig transform to ScenarioRealisation

    if isinstance(scenario, ScenarioConfig):
        scenario = scenario.create_realisation(start_date, end_date, resolution)

    assert isinstance(scenario, ScenarioRealisation), 'Realisation must be of type ' \
                                                      'ScenarioRealisation or ScenarioConfig.'

    # empty log file
    with open('log.log', 'w'):
        pass
    if log:
        logging.basicConfig(filename='log.log', level=logging.INFO)
    # get list with all time_steps as datetime.datetime
    time_steps = create_time_steps(scenario.start_date, scenario.end_date, scenario.resolution)
    # Initialize waiting queue for cars at the infrastructure
    waiting_queue = WaitingQueue(maxsize=scenario.queue_length)
    # copy charging_events from scenario
    charging_events = scenario.charging_events
    # set up infrastructure and get all charging points
    free_cps = set(set_up_infrastructure(scenario.infrastructure))
    busy_cps = set()
    # Rejections counter
    counter_rejections = 0
    # Charging times tracker
    charging_periods = {}
    # Opening hours
    opening_hours = scenario.opening_hours

    charging_event_counter = 0

    # ---------------------  Main Loop  ---------------------------
    # loop over every time step
    total_time_steps = len(time_steps)
    for time_step_pos in range(len(time_steps)):
        time_step = time_steps[time_step_pos]
        if log:
            logging.info(' %s', time_step)
        if time_step_pos % (int(0.05 * total_time_steps)) == 1:
            yield time_step_pos/total_time_steps
        if opening_hours is None:
            within_opening_hours = True
        else:
            cur_time_hours = time_step.hour + time_step.minute/60 + time_step.second/60/60
            within_opening_hours = opening_hours[0] <= cur_time_hours <= opening_hours[1]

        # check if cars must be disconnected, if yes immediately connect car from queue if possible
        update_queue(waiting_queue, time_step, scenario.disconnect_by_time,
                     within_opening_hours, log)

        update_cps(free_cps, busy_cps, waiting_queue, time_step,
                   scenario.disconnect_by_time, within_opening_hours, log)

        # in case of multiple charging events in the same time step: handle one after the other
        while len(charging_events) > charging_event_counter and \
                time_step == charging_events[charging_event_counter].arrival_time:

            current_charging_event = charging_events[charging_event_counter]
            waiting_queue, counter_rejections = handle_car_arrival(
                                                free_cps, busy_cps,
                                                current_charging_event, waiting_queue,
                                                counter_rejections, within_opening_hours, log)
            charging_event_counter += 1

        # assign power
        assign_power = scenario.scheduling_policy.schedule(scenario, free_cps,
                                                           busy_cps, time_step_pos)
        if len(busy_cps) > 0:
            charging_periods = update_last_charged(charging_periods, assign_power['cps'], time_step)

        charge_connected_vehicles(assign_power['cps'], busy_cps, scenario.resolution, log)
        charge_storage(assign_power, scenario.transformer_preload[time_step_pos],
                       scenario.resolution)
        results.store_power_charging_points(assign_power['cps'], time_step_pos,
                                            time_step_pos == total_time_steps-1)
        results.store_power_storage_systems(assign_power['storage'], time_step_pos,
                                            time_step_pos == total_time_steps-1)

    results.counter_rejections = counter_rejections
    results.charging_periods = charging_periods

if __name__ == '__main__':

    start_date = datetime.datetime(2020, 1, 1)  # '2020-01-01 00:00:00'
    end_date = datetime.datetime(2020, 1, 7, 23, 59)
    resolution = datetime.timedelta(hours=1) # '01:0:0'
    time_params = (start_date, end_date, resolution)
    # time_params = (start_date, end_date, resolution)
    num_charging_events = 500
    #
    arrival_distribution = [0 for x in range(84)] #[np.random.uniform(0, 1) for x in range(168)]
    arrival_distribution[4] = 1
    arrival_distribution[5] = 1
    #
    queue_length = 2
    infrastructure = {'transformers': [{'id': 'transformer1', 'max_power': 100, 'min_power': 10, 'charging_stations': [{'id': 'cs1', 'max_power': 10, 'min_power': 1, 'charging_points': [{'id': 'cp1', 'max_power': 5, 'min_power': 0.5}, {'id': 'cp2', 'max_power': 5, 'min_power': 0.5}]}, {'id': 'cs2', 'max_power': 10, 'min_power': 1, 'charging_points': [{'id': 'cp3', 'max_power': 5, 'min_power': 0.5}, {'id': 'cp4', 'max_power': 5, 'min_power': 0.5}]}]}]}
    disconnect_by_time = True
    # scheduling_policy = Uncontrolled()
    #
    # conf = ElvisConfig(arrival_distribution, None, None, infrastructure, None, scheduling_policy,
    #                    None, time_params, num_charging_events, queue_length, disconnect_by_time)
    # simulate(conf)

    config = ScenarioConfig()
    config.with_scheduling_policy('UC')
    config.with_std_deviation_soc(0.3)
    config.with_mean_soc(0.4)
    #scenario.with_scheduling_policy(FCFS())
    config.with_infrastructure(infrastructure)
    config.with_disconnect_by_time(disconnect_by_time)
    config.with_queue_length(queue_length)
    config.with_num_charging_events(num_charging_events)
    config.with_mean_park(4)
    config.with_std_deviation_park(1)
    kwargs = {'brand': 'VW', 'model': 'e-Golf', 'probability': 1, 'battery': {'capacity': 35.8,
              'min_charge_power': 0, 'max_charge_power': 150, 'efficiency': 1}}
    config.with_vehicle_types(**kwargs)
    kwargs = {'brand': 'VW', 'model': 'Up', 'probability': 1, 'battery': {'capacity': 35.8,
                                                                              'min_charge_power': 0,
                                                                              'max_charge_power': 150,
                                                                              'efficiency': 1}}
    config.with_vehicle_types(**kwargs)
    config.with_arrival_distribution(arrival_distribution)
    import random
    config.with_transformer_preload([0] * 10000)

    result = simulate(config, start_date, end_date, resolution)
    print(result.power_charging_points)
    # load_profile = result.aggregate_load_profile(scenario.num_simulation_steps())
    # print(list(zip(load_profile, create_time_steps(scenario.start_date, scenario.end_date, scenario.resolution))))




"""This module contains all classes necessary to configure scenarios that can then be simulated
by Elvis."""

import logging
import warnings
import datetime
import pandas as pd
import yaml
import gzip
import json

import elvis.sched.schedulers as schedulers
from elvis.charging_event_generator import create_charging_events_from_weekly_distribution as \
    events_from_week_arr_dist
from elvis.charging_event_generator import create_charging_events_from_gmm as events_from_gmm
from elvis.utility.elvis_general import create_time_steps, num_time_steps, transform_data, \
    adjust_resolution, repeat_data
from elvis.battery import EVBattery
from elvis.vehicle import ElectricVehicle
from elvis.charging_event import ChargingEvent
from elvis.set_up_infrastructure import wallbox_infrastructure


class ScenarioConfig:
    """Describes a scenario defined by its stochastic distributions and hardware parameters."""
    def __init__(self, **kwargs):
        # Time series data
        self.emissions_scenario = None
        self.emissions_scenario_res_data = None
        self.emissions_scenario_repeat = None
        self.emissions_scenario_col_pos = 0
        self.renewables_scenario = None

        self.transformer_preload = None
        self.transformer_preload_res_data = None
        self.transformer_preload_repeat = False
        self.transformer_preload_col_pos = 0

        # Event parameters
        self.sample_method = 'independent_normal_dist'  # or GMM
        self.arrival_distribution = None
        self.vehicle_types = []
        self.mean_park = None
        self.std_deviation_park = None
        self.mean_soc = None
        self.std_deviation_soc = None
        self.max_parking_time = 24
        self.num_charging_events = None
        self.charging_events = None
        self.gmm_means = None
        self.gmm_weights = None
        self.gmm_covariances = None

        # Infrastructure (behaviour)
        self.infrastructure = None
        self.queue_length = None
        self.disconnect_by_time = None
        self.opening_hours = None
        self.scheduling_policy = None
        self.df_charging_period = None

        if 'emissions_scenario' in kwargs:
            self.with_emissions_scenario(kwargs['emissions_scenario'])
        if 'renewables_scenario' in kwargs:
            self.renewables_scenario = kwargs['renewables_scenario']
        if 'transformer_preload' in kwargs:
            self.transformer_preload = kwargs['transformer_preload']

        if 'vehicle_types' in kwargs:
            self.vehicle_types = kwargs['vehicle_types']

        if 'opening_hours' in kwargs:
            self.with_opening_hours(kwargs['opening_hours'])

        if 'sample_method' in kwargs:
            self.sample_method = kwargs['sample_method']

        if 'charging_events' in kwargs:
            self.charging_events = kwargs['charging_events']
        if 'arrival_distribution' in kwargs:
            self.with_arrival_distribution(kwargs['arrival_distribution'])
        if 'infrastructure' in kwargs:
            self.with_infrastructure(kwargs['infrastructure'])
        if 'scheduling_policy' in kwargs:
            self.with_scheduling_policy(kwargs['scheduling_policy'])
        if isinstance(self.scheduling_policy, schedulers.DiscriminationFree):
            if 'df_charging_period' in kwargs:
                self.with_df_charging_period(kwargs['df_charging_period'])
            else:
                self.with_df_charging_period(datetime.timedelta(minutes=15))
        if 'mean_park' in kwargs:
            self.with_mean_park(kwargs['mean_park'])
        if 'std_deviation_park' in kwargs:
            self.with_std_deviation_park(kwargs['std_deviation_park'])
        if 'mean_soc' in kwargs:
            self.with_mean_soc(kwargs['mean_soc'])
        if 'std_deviation_soc' in kwargs:
            self.with_std_deviation_soc(kwargs['std_deviation_soc'])

        # Gaussian Mixture Model params
        if 'gmm_means' in kwargs:
            self.gmm_means = kwargs['gmm_means']
        if 'gmm_weights' in kwargs:
            self.gmm_weights = kwargs['gmm_weights']
        if 'gmm_covariances' in kwargs:
            self.gmm_covariances = kwargs['gmm_covariances']

        if 'max_parking_time' in kwargs:
            self.max_parking_time = kwargs['max_parking_time']
        if 'num_charging_events' in kwargs:
            self.with_num_charging_events(kwargs['num_charging_events'])
        if 'queue_length' in kwargs:
            self.with_queue_length(kwargs['queue_length'])
        if 'disconnect_by_time' in kwargs:
            self.with_disconnect_by_time(kwargs['disconnect_by_time'])
        if 'transformer_preload_res_data' in kwargs:
            self.transformer_preload_res_data = kwargs['transformer_preload_res_data']
        if 'transformer_preload_repeat' in kwargs:
            self.transformer_preload_repeat = kwargs['transformer_preload_repeat']
        if 'transformer_preload_col_pos' in kwargs:
            self.transformer_preload_col_pos = kwargs['transformer_preload_col_pos']
        if 'emissions_scenario_res_data' in kwargs:
            self.emissions_scenario_res_data = kwargs['emissions_scenario_res_data']
        if 'emissions_scenario_repeat' in kwargs:
            self.emissions_scenario_repeat = kwargs['emissions_scenario_repeat']
        if 'emissions_scenario_col_pos' in kwargs:
            self.emissions_scenario_col_pos = kwargs['emissions_scenario_col_pos']

    def __str__(self):
        printout = ''
        if self.vehicle_types is None:
            printout += str('Vehicle types: None\n')
        else:
            printout = 'Vehicle types: ' + str(vt + '\n' for vt in self.vehicle_types)

        printout += str('Mean parking time: ' + str(self.mean_park) + '\n')
        printout += str('Std deviation of parking time: ' + str(self.std_deviation_park) + '\n')
        printout += str('Mean value of the SOC distribution: ' + str(self.mean_soc) + '\n')
        printout += str('Std deviation of the SOC distribution: ' +
                        str(self.std_deviation_soc) + '\n')
        printout += str('Max parking time: ' +
                        str(self.max_parking_time) + '\n')
        printout += str('Number of charging events per week: ' +
                        str(self.num_charging_events) + '\n')

        if self.disconnect_by_time is True:
            printout += str('Vehicles are disconnected only depending on their parking time\n')
        else:
            printout += str('Vehicles are disconnected depending on their SOC and their parking'
                            'parking time (what ever comes first\n')
        if self.queue_length is None:
            printout += str('No queue is considered.\n')
        else:
            printout += str('Queue length: ' + str(self.queue_length) + '\n')

        printout += str('Opening hours: ' + str(self.opening_hours) + '\n')
        printout += str('Scheduling policy: ' + str(self.scheduling_policy) + '\n')

        return printout

    def to_dict(self):
        dictionary = self.__dict__.copy()
        if self.scheduling_policy is not None:
            dictionary['scheduling_policy'] = str(self.scheduling_policy)
        if self.charging_events is not None:
            dictionary['charging_events'] = [ce.to_dict() for ce in self.charging_events]

        # TODO what if transformer preload is not a list
        dictionary['transformer_preload'] = self.transformer_preload

        if len(self.vehicle_types) > 0:
            dictionary['vehicle_types'] = [vehicle.to_dict() for vehicle in self.vehicle_types]

        return dictionary

    @staticmethod
    def from_dict(dictionary):

        assert type(dictionary) is dict, 'Input of wrong type: ' + str(type(dictionary))

        config = ScenarioConfig()
        config.with_infrastructure(dictionary['infrastructure'])
        config.with_scheduling_policy(dictionary['scheduling_policy'])
        if 'mean_park' in dictionary:
            config.with_mean_park(dictionary['mean_park'])
        if 'std_deviation_park' in dictionary:
            config.with_std_deviation_park(dictionary['std_deviation_park'])
        if 'mean_soc' in dictionary:
            config.with_mean_soc(dictionary['mean_soc'])
        if 'std_deviation_soc' in dictionary:
            config.with_std_deviation_soc(dictionary['std_deviation_soc'])
        config.with_num_charging_events(dictionary['num_charging_events'])
        config.with_queue_length(dictionary['queue_length'])
        config.with_disconnect_by_time(dictionary['disconnect_by_time'])

        if isinstance(config.scheduling_policy, schedulers.DiscriminationFree):
            if 'df_charging_period' in dictionary:
                config.with_df_charging_period(dictionary['df_charging_period'])
            else:
                config.with_df_charging_period(datetime.timedelta(minutes=15))

        if 'sample_method' in dictionary:
            config.sample_method = dictionary['sample_method']
        if 'max_parking_time' in dictionary:
            config.max_parking_time = dictionary['max_parking_time']

        if 'opening_hours' in dictionary:
            config.with_opening_hours(dictionary['opening_hours'])

        if 'transformer_preload' in dictionary:
            config.transformer_preload = dictionary['transformer_preload']

        config.with_vehicle_types(vehicle_types=dictionary['vehicle_types'])
        if 'arrival_distribution' in dictionary:
            config.with_arrival_distribution(dictionary['arrival_distribution'])
        if 'charging_events' in dictionary:
            config.charging_events = dictionary['charging_events']

        if 'gmm_means' in dictionary:
            config.gmm_means = dictionary['gmm_means']
        if 'gmm_weights' in dictionary:
            config.gmm_weights = dictionary['gmm_weights']
        if 'gmm_covariances' in dictionary:
            config.gmm_covariances = dictionary['gmm_covariances']

        if 'transformer_preload_res_data' in dictionary:
            config.transformer_preload_res_data = dictionary['transformer_preload_res_data']
        if 'transformer_preload_repeat' in dictionary:
            config.transformer_preload_repeat = dictionary['transformer_preload_repeat']
        if 'transformer_preload_col_pos' in dictionary:
            config.transformer_preload_col_pos = dictionary['transformer_preload_col_pos']
        if 'emissions_scenario_res_data' in dictionary:
            config.emissions_scenario_res_data = dictionary['emissions_scenario_res_data']
        if 'emissions_scenario_repeat' in dictionary:
            config.emissions_scenario_repeat = dictionary['emissions_scenario_repeat']
        if 'emissions_scenario_col_pos' in dictionary:
            config.emissions_scenario_col_pos = dictionary['emissions_scenario_col_pos']

        return config

    def to_yaml(self, yaml_file):
        data = self.to_dict()
        with open(yaml_file, 'w') as file:
            yaml.dump(data, file, default_flow_style=None)
        return

    @staticmethod
    def from_yaml(yaml_str):
        """Create an instance of ScenarioConfig from a yaml str.

        Args:
            yaml_str: (dict): Import from yaml file.
            """
        assert type(yaml_str) == dict, 'The data specified in the yaml for creating an instance ' \
                                       'Scenario scenario must be a dict.'

        return ScenarioConfig.from_dict(yaml_str)

    def create_realisation(self, start_date, end_date, resolution):
        """Creates a realisation of self given the required time parameters.

        Args:
            start_date: (:obj: `datetime.datetime`): First time stamp.
            end_date: (:obj: `datetime.datetime`): Upper bound for time stamps.
            resolution: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.

        Returns:
            realisation: (:obj: `elvis.scenario.ScenarioRealisation`): Scenario realisation.
        """

        time_params = {'start_date': start_date, 'end_date': end_date, 'resolution': resolution}
        realisation = ScenarioRealisation(config=self, **time_params)
        return realisation

    def with_arrival_distribution(self, arrival_distribution):
        # TODO: Add pandas
        assert type(arrival_distribution) is list, 'Arrival distribution must be of type list or ' \
                                                   'pandas DataFrame.'

        msg_invalid_value_type = "Arrival distribution should be of type: pandas DataFrame or a " \
                                 "list containing float or int."
        # Check if all values in list are either float or int
        if arrival_distribution is list:
            assert all(isinstance(x, (float, int)) for x in arrival_distribution), \
                msg_invalid_value_type
        else:
            NotImplementedError()

        self.arrival_distribution = arrival_distribution

        return

    # Outdated
    def with_charging_events(self, charging_events):
        """Update the arrival distribution to use.
        """
        assert isinstance(charging_events, list), 'charging_events must be of type list'
        assert all(isinstance(x, ChargingEvent) for x in charging_events), \
            'All elements of charging_events must be of type ChargingEvent'
        self.charging_events = charging_events

        return self

    def with_emissions_scenario(self, emissions_scenario, col_pos=0, res_data=None, repeat=False):
        """Assigns values regarding the emissions scenario to config.

        Args:
            emissions_scenario: Either of type int/float, list, or pandas DataFrame/Series.
            col_pos: (int): If emissions scenario is of type pandas DataFrame: col_pos refers to
                the column the emissions scenario is stored.
            res_data: (datetime.timedelta): If emissions scenario is of type list and does not align
                with the simulation period: res_data states the time difference in between two
                adjacent values of emissions scenario.
            repeat: (bool): If emissions scenario is of type list and does not align with the
                simulation period: repeat=True indicates that the list shall be repeated until
                the whole simulation period is covered.
        """
        if isinstance(emissions_scenario, (pd.Series, pd.DataFrame)):
            self.emissions_scenario = emissions_scenario
        else:
            assert isinstance(emissions_scenario, (int, float, list)), \
                'Emissions scenario must be of type list or pandas DataFrame.'

            msg_invalid_value_type = "Emissions scenario should be of type: pandas DataFrame or" \
                                     " a list containing float or int."
            # Check if all values in list are either float or int
            if type(emissions_scenario) is list:
                assert all(isinstance(x, (float, int)) for x in emissions_scenario), \
                    msg_invalid_value_type

            self.emissions_scenario = emissions_scenario

        if res_data is not None:
            assert isinstance(res_data, (str, datetime.timedelta))
            if type(res_data) is str:
                try:
                    date = datetime.datetime.strptime(res_data, '%H:%M:%S')
                    self.emissions_scenario_res_data = datetime.timedelta(hours=date.hour,
                                                                          minutes=date.minute,
                                                                          seconds=date.second)
                except ValueError:
                    try:
                        seconds = pd.Timedelta(res_data).total_seconds()
                        self.emissions_scenario_res_data = datetime.timedelta(seconds=seconds)
                    except ValueError:
                        print('Incorrect timedelta format for resolution pls use: %H:%M:%S or '
                              'a pandas conform timedelta format.')
            # datetime.timedelta
            else:
                self.emissions_scenario_res_data = res_data

        if repeat is not False:
            assert type(repeat) is bool, 'Repeat can only be True or False.'
            self.emissions_scenario = True

        if col_pos != 0:
            assert isinstance(col_pos, int), 'Column position (col_pos) must be a positive integer'
            assert col_pos >= 0, 'Column position (col_pos) must be a positive integer'
            self.emissions_scenario_col_pos = col_pos
        return self

    def with_renewables_scenario(self, renewables_scenario):
        """Update the renewable energy scenario to use."""

        self.renewables_scenario = renewables_scenario
        return self

    def with_transformer_preload(self, transformer_preload, col_pos=0, res_data=None, repeat=False):
        """Assigns values regarding the transformer preload to config.

        Args:
            transformer_preload: Either of type int/float, list, or pandas DataFrame/Series.
            col_pos: (int): If transformer_preload is of type pandas DataFrame: col_pos refers to
                the column the transformer preload is stored.
            res_data: (datetime.timedelta): If transfomer preload is of type list and does not align
                with the simulation period: res_data states the time difference in between two
                adjacent values of transformer preload.
            repeat: (bool): If transformer preload is of type list and does not align with the
                simulation period: repeat=True indicates that the list shall be repeated until
                the whole simulation period is covered.
        """
        if isinstance(transformer_preload, (pd.Series, pd.DataFrame)):
            self.transformer_preload = transformer_preload
        else:
            assert isinstance(transformer_preload, (int, float, list)), \
                'Transformer preload must be of type list or pandas DataFrame.'

            msg_invalid_value_type = "Transformer preload should be of type: pandas DataFrame or" \
                                     " a list containing float or int."
            # Check if all values in list are either float or int
            if type(transformer_preload) is list:
                assert all(isinstance(x, (float, int)) for x in transformer_preload), \
                    msg_invalid_value_type

            self.transformer_preload = transformer_preload

        if res_data is not None:
            assert isinstance(res_data, (str, datetime.timedelta))
            if type(res_data) is str:
                try:
                    date = datetime.datetime.strptime(res_data, '%H:%M:%S')
                    self.transformer_preload_res_data = datetime.timedelta(hours=date.hour,
                                                                           minutes=date.minute,
                                                                           seconds=date.second)
                except ValueError:
                    try:
                        seconds = pd.Timedelta(res_data).total_seconds()
                        self.transformer_preload_res_data = datetime.timedelta(seconds=seconds)
                    except ValueError:
                        print('Incorrect timedelta format for resolution pls use: %H:%M:%S or '
                              'a pandas conform timedelta format.')
            # datetime.timedelta
            else:
                self.transformer_preload_res_data = res_data

        if repeat is not False:
            assert type(repeat) is bool, 'Repeat can only be True or False.'
            self.transformer_preload_repeat = True

        if col_pos != 0:
            assert isinstance(col_pos, int), 'Column position (col_pos) must be a positive integer'
            assert col_pos > 0, 'Column position (col_pos) must be a positive integer'
            self.transformer_preload_col_pos = col_pos
        return self

    def with_scheduling_policy(self, scheduling_policy_input):
        """Update the scheduling policy to use.
        Default: :obj: `elvis.sched.schedulers.Uncontrolled`.
        Use default if input not a str or str can not be matched.

        Args:
            scheduling_policy_input: Either str containing name of the scheduling policy to be used.
                Or instance of :obj: `elvis.sched.schedulers.SchedulingPolicy`.
        """
        # set default
        scheduling_policy = schedulers.Uncontrolled()

        # if input is already instance of Scheduling Policy assign
        if isinstance(scheduling_policy_input, schedulers.SchedulingPolicy):
            self.scheduling_policy = scheduling_policy_input
            return self

        # ensure input is str. If not return default
        if type(scheduling_policy_input) is not str:
            logging.error('Scheduling policy should be of type str or an instance of '
                          'SchedulingPolicy. The uncontrolled strategy has been used as a default.')
            self.scheduling_policy = scheduling_policy
            return self

        # Match string
        if scheduling_policy_input in ('Uncontrolled', 'UC', 'Uc', 'uc'):
            self.scheduling_policy = schedulers.Uncontrolled()

        elif scheduling_policy_input in ('Discrimination Free', 'DF', 'df'):
            self.scheduling_policy = schedulers.DiscriminationFree()

        elif scheduling_policy_input == 'FCFS':
            self.scheduling_policy = schedulers.FCFS()

        elif scheduling_policy_input in ('With Storage', 'ws', 'WS'):
            self.scheduling_policy = schedulers.WithStorage()

        elif scheduling_policy_input in ('Optimized', 'opt', 'OPT'):
            self.scheduling_policy = schedulers.Optimized()

        # invalid str use default: Uncontrolled
        else:
            logging.error('"%s" can not be matched to any existing scheduling policy.'
                          'Please use: "Uncontrolled", "Discrimination Free", "FCFS", '
                          '"With Storage" or "Optimized". '
                          'Uncontrolled is the default value and has been used for the simulation.',
                          str(scheduling_policy_input))

            self.scheduling_policy = schedulers.Uncontrolled()
        return self

    def with_infrastructure(self, infrastructure=None, **kwargs):
        """Update the infrastructure to use.
        Args:
            kwargs.num_cp: (int): Number of charging points.
            kwargs.num_cp_per_cs: (int) Number of charging points per charging station.
            kwargs.power_cp: (int or float): (optional) Max power per charging point.
            kwargs.power_cs: (int or float): (optional) Max power of the charging station.
            kwargs.power_transformer: (int or float): (optional) Max power of the transformer.
            kwargs.min_power_cp: (int or float): (optional) Minimum power (if not 0) for the
                charging point.
            kwargs.min_power_cs: (int or float): (optional) Minimum power (if not 0) for the
                charging station.
            kwargs.min_power_transformer: (int or float) : (optional) Minimum power (if not 0) for
                the charging station.
        """
        if infrastructure is not None:
            assert type(infrastructure) is dict
            self.infrastructure = infrastructure
        else:
            err_msg = '%s is a necessary key to initialise an elvis infrastructure as a wallbox ' \
                      'infrastructure.'
            keys = kwargs.keys()
            assert 'num_cp' in keys, (err_msg, 'num_cp')
            assert 'power_cp' in keys, (err_msg, 'power_cp')

            num_cp = kwargs['num_cp']
            power_cp = kwargs['power_cp']

            if 'num_cp_per_cs' in keys:
                num_cp_per_cs = kwargs['num_cp_per_cs']
            else:
                num_cp_per_cs = 1
            if 'power_cs' in keys:
                power_cs = kwargs['power_cs']
            else:
                power_cs = None
            if 'power_transformer' in keys:
                power_transformer = kwargs['power_transformer']
            else:
                power_transformer = None
            if 'min_power_cp' in keys:
                min_power_cp = kwargs['min_power_cp']
            else:
                min_power_cp = 0
            if 'min_power_cs' in keys:
                min_power_cs = kwargs['min_power_cs']
            else:
                min_power_cs = 0
            if 'min_power_transformer' in keys:
                min_power_transformer = kwargs['min_power_transformer']
            else:
                min_power_transformer = 0

            self.infrastructure = wallbox_infrastructure(
                num_cp, power_cp, num_cp_per_cs, power_cs, power_transformer, min_power_cp,
                min_power_cs, min_power_transformer)

        return self

    def with_vehicle_types(self, vehicle_types=None, **kwargs):
        """Update the vehicle types to use."""

        if vehicle_types is not None:
            for vehicle_type in vehicle_types:
                if isinstance(vehicle_type, ElectricVehicle):
                    self.add_vehicle_types(vehicle_type=vehicle_type)
                elif isinstance(vehicle_type, dict):
                    self.add_vehicle_types(**vehicle_type)
                else:
                    # TODO: List relevant information.
                    raise TypeError('Using with_vehicle_types with a list: All list entries '
                                    'must either be of type ElectricVehicle or dict, containing'
                                    'all relevant information.')

            return self
        else:
            self.add_vehicle_types(**kwargs)
            return self

    def with_df_charging_period(self, charging_period):

        assert isinstance(charging_period, (str, datetime.timedelta)), (
            'Charging period must either be a str in format: %H:%M:%S or an instance '
            'of datetime.timedelta')

        # str
        if type(charging_period) is str:
            try:
                date = datetime.datetime.strptime(charging_period, '%H:%M:%S')
                self.df_charging_period = datetime.timedelta(hours=date.hour,
                                                             minutes=date.minute,
                                                             seconds=date.second)
            except ValueError:
                try:
                    seconds = pd.Timedelta(charging_period).total_seconds()
                    self.df_charging_period = datetime.timedelta(seconds=seconds)
                except ValueError:
                    print('Incorrect timedelta format for resolution pls use: %H:%M:%S or '
                          'a pandas conform timedelta format.')
        # datetime.timedelta
        else:
            self.df_charging_period = charging_period

        return self

    def add_vehicle_types(self, vehicle_type=None, **kwargs):
        """Add a supported vehicle type to this configuration or a list of vehicle types.
            If no instance of vehicle type is passed an instance of vehicle_type is created if
            kwargs contains the necessary keys (see below).
            If vehicle_type is passed:
                - vehicle_type: (:obj: `vehicle.ElectricVehicle`

            If battery instance is passed kwargs has to to contain:
                - brand: (str): Brand of the vehicle.
                - model: (str): Model of the vehicle.
                - battery: (:obj: `battery.EVBattery`): Battery instance
            If no battery is passed and shall be initialized:
                - brand: (str): Brand of the vehicle.
                - model: (str): Model of the vehicle.
                - capacity: (float): Capacity of the battery in kWh.
                - max_charge_power: (float): Max power in kW.
                - min_charge_power: (float): Min power in kW.
                - efficiency: (float): [0, 1]

            Raises:
                AssertionError:
                    - If vehicle_type is of type list and at least one of the entries is not of
                        type :obj: `ElectricVehicle`.
                    - If vehicle type is not a list and not of type :obj: `ElectricVehicle`.
                    - If kwargs are passed and the keys do not contain the variable names listed
                        above.
        """
        # if list with multiple vehicle_type instances is passed add multiple
        if type(vehicle_type) is list:
            for vehicle in vehicle_type:
                assert isinstance(vehicle, ElectricVehicle)
                self.vehicle_types.append(vehicle)
                return
        # if an instance of vehicle type is passed assign
        elif vehicle_type is not None:
            assert isinstance(vehicle_type, ElectricVehicle)
            self.vehicle_types.append(vehicle_type)
            return

        assert 'brand' in kwargs
        assert 'model' in kwargs
        assert 'probability' in kwargs
        assert 'battery' in kwargs

        brand = str(kwargs['brand'])
        model = str(kwargs['model'])
        probability = kwargs['probability']

        # if all fields of ElectricVehicle are passed and an instance of EVBattery is already made
        if isinstance(kwargs['battery'], EVBattery):
            battery = kwargs['battery']
            self.vehicle_types.append(ElectricVehicle(brand, model, battery, probability))
            return

        # if there is no instance of EVBattery check if all parameters are passed.
        assert 'capacity' in kwargs['battery']
        assert 'max_charge_power' in kwargs['battery']
        assert 'min_charge_power' in kwargs['battery']
        assert 'efficiency' in kwargs['battery']

        capacity = kwargs['battery']['capacity']
        max_charge_power = kwargs['battery']['max_charge_power']
        min_charge_power = kwargs['battery']['min_charge_power']
        efficiency = kwargs['battery']['efficiency']

        # Power degradations when battery reaches certain SOC level shall be considered
        if 'start_power_degradation' and 'max_degradation_level' in kwargs['battery']:
            start_power_degradation = kwargs['battery']['start_power_degradation']
            max_degradation_level = kwargs['battery']['max_degradation_level']

            battery = EVBattery(capacity=capacity,
                                max_charge_power=max_charge_power,
                                min_charge_power=min_charge_power,
                                efficiency=efficiency,
                                start_power_degradation=start_power_degradation,
                                max_degradation_level=max_degradation_level)
        # No SOC-dependent power degradations considered
        else:
            battery = EVBattery(capacity=capacity, max_charge_power=max_charge_power,
                                min_charge_power=min_charge_power, efficiency=efficiency, )

        # get instance of ElectricVehicle with initialized battery
        self.vehicle_types.append(ElectricVehicle(brand, model, battery, probability))
        return self

    def with_opening_hours(self, opening_hours):
        """Update the opening hours to use."""

        if opening_hours is None:
            self.opening_hours = opening_hours
            return self

        assert isinstance(opening_hours, tuple), 'Opening hours is expected to be a tuple.'
        assert len(opening_hours) == 2, 'Opening hours is expected to be a tuple with 2 values'
        _open = opening_hours[0]
        _close = opening_hours[1]
        assert isinstance(_open, (float, int)), 'Values in opening hours must be of type int or ' \
                                                'float representing the hours of the day.'
        assert isinstance(_open, (float, int)), 'Values in opening hours must be of type int or ' \
                                                'float representing the hours of the day.'

        assert _open <= _close, 'The first value (opening hour) is expected to be smaller than ' \
                                'the 2nd value (closing hour).'
        assert _close <= 24, 'The last value(closing hour) is expected to be smaller or equal to 24'
        assert _open >= 0, 'The first value(opening hour) is expected to be bigger or equal to 0'

        self.opening_hours = opening_hours
        return self

    def with_num_charging_events(self, num_charging_events):
        """Update the number of charging events to use."""
        # TODO: Can be float?
        assert type(num_charging_events) is int
        self.num_charging_events = num_charging_events
        return self

    def with_queue_length(self, queue_length):
        """Update maximal length of queue."""
        assert type(queue_length) is int
        self.queue_length = queue_length
        return self

    def with_disconnect_by_time(self, disconnect_by_time):
        """Update decision variable on how to disconnect cars."""
        assert type(disconnect_by_time) is bool
        self.disconnect_by_time = disconnect_by_time
        return self

    def with_mean_park(self, mean_park):
        """Update decision variable on how to disconnect cars."""
        assert isinstance(mean_park, (int, float))
        assert 0 < mean_park < 24, "The mean parking time should be in between 0 and 24 hours."
        self.mean_park = mean_park
        return self

    def with_std_deviation_park(self, std_deviation_park):
        """Update decision variable on how to disconnect cars."""
        assert isinstance(std_deviation_park, (int, float)), 'The mean of the std deviation of ' \
                                                             'the parking time must be of type ' \
                                                             'float.'

        assert 0 <= std_deviation_park
        self.std_deviation_park = std_deviation_park
        return self

    def with_max_parking_time(self, max_parking_time):
        """Update decision variable on how to disconnect cars."""
        assert isinstance(max_parking_time, (int, float)), 'The mean of the std deviation of ' \
                                                           'the parking time must be of type ' \
                                                           'float.'

        assert 0 <= max_parking_time
        self.max_parking_time = max_parking_time
        return self

    def with_mean_soc(self, mean_soc):
        """Update decision variable on how to disconnect cars."""
        assert isinstance(mean_soc, (int, float)), 'The mean of the SOC must be of type float.'
        assert 0 < mean_soc < 1, "The mean parking time should be in between 0 and 24 hours."
        self.mean_soc = mean_soc
        return self

    def with_std_deviation_soc(self, std_deviation_soc):
        """Update decision variable on how to disconnect cars."""
        assert isinstance(std_deviation_soc, (int, float)), 'The std deviation of the SOC must' \
                                                            'be of type int or float.'
        assert 0 <= std_deviation_soc, 'The std deviation of the SOC must be bigger than 0.'
        self.std_deviation_soc = std_deviation_soc
        return self


class ScenarioRealisation:
    """Describes a realisation based on the stochasticity of a ScenarioConfig."""

    @staticmethod
    def check_input(**kwargs):
        """Check if all necessary keys are in kwargs of __init__."""
        keys_necessary = ['emissions_scenario', 'renewables_scenario', 'opening_hours',
                          'infrastructure', 'scheduling_policy', 'queue_length',
                          'disconnect_by_time', 'start_date', 'end_date', 'resolution']

        for key in keys_necessary:
            assert key in kwargs, str(key) + ' is missing as an input to create a ' \
                                             'ScenarioRealisation.'
        return

    def __init__(self, config=None, **kwargs):
        """Can be either called with an instance of ScenarioConfig or a dict containing all
        necessary parameters."""

        self.charging_events = None
        self.emissions_scenario = None
        self.transformer_preload = None
        self.scheduling_policy = None

        # Time parameters
        try:
            # Start date
            assert isinstance(kwargs['start_date'], (str, datetime.datetime)), (
                'start_date must either be an isoformat compatible datetime str or an instance '
                'of datetime.datetime')
            # str
            if type(kwargs['start_date']) is str:
                try:
                    self.start_date = datetime.datetime.fromisoformat(kwargs['start_date'])
                except ValueError:
                    print('Incorrect date format for start_date pls use: %y-%m-%d %H:%M:%S')
            # datetime.datetime
            else:
                self.start_date = kwargs['start_date']

            # End Date
            assert isinstance(kwargs['end_date'], (str, datetime.datetime)), (
                'end_date must either be an isoformat compatible datetime str or an instance '
                'of datetime.datetime')

            # str
            if type(kwargs['end_date']) is str:
                try:
                    self.end_date = datetime.datetime.fromisoformat(kwargs['end_date'])
                except ValueError:
                    print('Incorrect date format for end_date pls use: %y-%m-%d %H:%M:%S')
            # datetime.datetime
            else:
                self.end_date = kwargs['end_date']

            # Resolution

            assert isinstance(kwargs['resolution'], (str, datetime.timedelta)), (
                'Resolution must either be a str in format: %H:%M:%S or an instance '
                'of datetime.timedelta')

            # str
            if type(kwargs['resolution']) is str:
                try:
                    date = datetime.datetime.strptime(kwargs['resolution'], '%H:%M:%S')
                    self.resolution = datetime.timedelta(hours=date.hour, minutes=date.minute,
                                                         seconds=date.second)
                except ValueError:
                    try:
                        seconds = pd.Timedelta(kwargs['resolution']).total_seconds()
                        self.resolution = datetime.timedelta(seconds=seconds)
                    except ValueError:
                        print('Incorrect timedelta format for resolution pls use: %H:%M:%S or '
                              'a pandas conform timedelta format.')
            # datetime.timedelta
            else:
                self.resolution = kwargs['resolution']

        # start_date, end_date or resolution not in kwargs
        except KeyError:
            print('Creating a ScenarioRealisation with a ScenarioConfig requires the'
                  'start_date, end_date and the resolution of the simulation to be passed '
                  'as key word arguments.')

        # With ScenarioConfig instance
        if isinstance(config, ScenarioConfig):
            self.with_emissions_scenario(config.emissions_scenario,
                                         self.start_date, self.end_date, self.resolution)
            self.renewables_scenario = config.emissions_scenario
            self.with_opening_hours(config.opening_hours)
            self.infrastructure = config.infrastructure
            self.with_scheduling_policy(config.scheduling_policy)
            self.sample_method = config.sample_method

            if isinstance(self.scheduling_policy, schedulers.DiscriminationFree):
                if config.df_charging_period is not None:
                    self.df_charging_period = config.df_charging_period
                else:
                    self.df_charging_period = datetime.timedelta(minutes=15)
            self.queue_length = config.queue_length
            self.disconnect_by_time = config.disconnect_by_time
            self.with_transformer_preload(config.transformer_preload,
                                          self.start_date, self.end_date, self.resolution,
                                          config.transformer_preload_col_pos,
                                          config.transformer_preload_res_data,
                                          config.transformer_preload_repeat)

            if config.charging_events is not None:
                self.charging_events = config.charging_events
            else:
                if self.sample_method is None or self.sample_method == 'independent_normal_dist':
                    self.transform_arrival_distribution()
                    time_steps = create_time_steps(self.start_date, self.end_date, self.resolution)
                    self.charging_events = events_from_week_arr_dist(
                        config.arrival_distribution, time_steps, config.num_charging_events,
                        config.mean_park, config.std_deviation_park, config.mean_soc,
                        config.std_deviation_soc, config.vehicle_types, config.max_parking_time)
                elif self.sample_method in ['GMM', 'gmm']:
                    time_steps = create_time_steps(self.start_date, self.end_date, self.resolution)
                    self.charging_events = events_from_gmm(
                        time_steps, config.num_charging_events, config.gmm_means,
                        config.gmm_weights, config.gmm_covariances, config.vehicle_types,
                        config.max_parking_time, config.mean_soc, config.std_deviation_soc)
                else:
                    ValueError('Events can only be sampled from independent weekly arrival '
                               'distributions ar a weekly GMM (gaussian mixture model)')

        else:
            self.check_input(**kwargs)
            self.with_emissions_scenario(kwargs['emissions_scenario'],
                                         self.start_date, self.end_date, self.resolution)
            self.renewables_scenario = kwargs['renewables_scenario']
            self.with_opening_hours(kwargs['opening_hours'])
            self.infrastructure = kwargs['infrastructure']
            self.with_scheduling_policy(kwargs['scheduling_policy'])
            if isinstance(self.scheduling_policy, schedulers.DiscriminationFree):
                if 'df_charging_period' in kwargs.keys():
                    self.with_df_charging_period(kwargs['df_chaging_period'])
                else:
                    self.df_charging_period = datetime.timedelta(minutes=15)
            self.queue_length = kwargs['queue_length']
            self.disconnect_by_time = kwargs['disconnect_by_time']
            self.start_date = kwargs['start_date']
            self.end_date = kwargs['end_date']
            self.resolution = kwargs['resolution']

            if 'charging_events' in kwargs:
                self.charging_events = [ChargingEvent.from_dict(**ce)
                                        for ce in kwargs['charging_events']]
            else:
                if self.sample_method is None or self.sample_method == 'independent_normal_dist':
                    time_steps = create_time_steps(self.start_date, self.end_date, self.resolution)
                    self.charging_events = events_from_week_arr_dist(
                        kwargs['arrival_distribution'], time_steps, kwargs['num_charging_events'],
                        kwargs['mean_park'], kwargs['std_deviation_park'], kwargs['mean_soc'],
                        kwargs['std_deviation_soc'], kwargs['vehicle_types'],
                        kwargs['max_parking_time'])
                elif self.sample_method in ['GMM', 'gmm']:
                    time_steps = create_time_steps(self.start_date, self.end_date, self.resolution)
                    self.charging_events = events_from_gmm(
                        time_steps, kwargs['num_charging_events'], kwargs['gmm_means'],
                        kwargs['gmm_weights'], kwargs['gmm_covariances'], kwargs['vehicle_types'],
                        kwargs['max_parking_time'], kwargs['mean_soc'], kwargs['std_deviation_soc'])
                else:
                    ValueError('Events can only be sampled from independent weekly arrival '
                               'distributions ar a weekly GMM (gaussian mixture model)')
            if 'transformer_preload' in kwargs:
                res_data = None
                repeat = None
                col_pos = 0
                if 'transformer_preload_res_data' in kwargs:
                    res_data = kwargs['transformer_preload_res_data']
                if 'transformer_preload_repeat' in kwargs:
                    repeat = kwargs['transformer_preload_repeat']
                if 'transformer_preload_col_pos' in kwargs:
                    col_pos = kwargs['transformer_preload_col_pos']

                self.with_transformer_preload(kwargs['transformer_preload'],
                                              self.start_date, self.end_date, self.resolution,
                                              col_pos, res_data, repeat)

        return

    def to_dict(self):
        dictionary = self.__dict__.copy()

        dictionary['scheduling_policy'] = str(self.scheduling_policy)
        # TODO: what if time series data is not a list
        dictionary['emissions_scenario'] = self.emissions_scenario
        dictionary['renewables_scenario'] = self.renewables_scenario
        dictionary['renewables_scenario'] = self.renewables_scenario

        dictionary['opening_hours'] = self.opening_hours
        dictionary['charging_events'] = [ce.to_dict(deep=True) for ce in self.charging_events]

        dictionary['start_date'] = str(self.start_date.isoformat())
        dictionary['end_date'] = str(self.end_date.isoformat())
        dictionary['resolution'] = str(self.resolution)

        if 'df_charging_period' in dictionary.keys():
            dictionary['df_charging_period'] = str(self.df_charging_period)

        return dictionary

    @staticmethod
    def from_dict(dictionary):
        realisation = ScenarioRealisation(**dictionary)
        return realisation

    # https://stackoverflow.com/questions/39450065/python-3-read-write-compressed-json-objects-from-to-gzip-file
    def save_to_disk(self, file_path):
        data = self.to_dict()

        json_str = json.dumps(data)
        json_bytes = json_str.encode('utf-8')

        with gzip.GzipFile(file_path, 'w') as fout:
            fout.write(json_bytes)

        return

    @staticmethod
    def from_disk(json_file_name):
        with gzip.GzipFile(json_file_name, 'r') as fin:
            json_bytes = fin.read()

        json_str = json_bytes.decode('utf-8')
        data = json.loads(json_str)

        return ScenarioRealisation.from_dict(data)

    def transform_arrival_distribution(self):
        pass

    def with_transformer_preload(self, transformer_preload, start_date, end_date, resolution,
                                 col_pos=0, res_data=None, repeat=False):
        """Update the renewable energy scenario to use.
        Simulation time parameter must be set before transformer preload can be initialised.

        Args:
            transformer_preload: Either of type int/float, list, or pandas Series/DataFrame.
            start_date: (:obj: `datetime.datetime`): First time stamp.
            end_date: (:obj: `datetime.datetime`): Upper bound for time stamps.
            resolution: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.
            col_pos: (int): If pandas DataFrame is passed indicates the column in which the
                emission values are stored.
            res_data: (datetime.timedelta): If list is passed that does not align with the
                simulation period, res_data displays the period inbetween two adjacent list entries.
            repeat: (bool): If list is passed that does not align with the
                simulation period and repeat is True the list is repeated until enough data points
                are available for the simulation period.

        """

        # Error messages
        msg_wrong_resolution_type = "If the transformer preload is passed as list and the" \
                                    "resolution is supposed to be of adjusted please" \
                                    "use type datetime.timedelta."
        msg_wrong_transformer_preload_type = "Transformer preload should be of type: " \
                                             "pandas DataFrame, list containing float or int," \
                                             " int, float."
        msg_alignement_unsuccessful = "Transformer preload could not be aligned to simulation" \
                                      "steps."
        msg_invalid_value_type = "Transformer preload should be of type: pandas DataFrame, " \
                                 "list containing float or int, int, float."
        msg_not_enough_data_points = "There are less values for the transformer preload than " \
                                     "simulation steps. Please adjust the data."

        # Check whether passed value is a DataFrame or a list
        if isinstance(transformer_preload, pd.DataFrame):  # DataFrame
            # Make sure length is aligned to simulation period and resolution
            # num_values = len(transformer_preload)
            transformer_preload = transformer_preload.iloc[:, col_pos]
            transformer_preload = transform_data(transformer_preload,
                                                 resolution, start_date, end_date)
        elif isinstance(transformer_preload, pd.Series):
            transformer_preload = transform_data(transformer_preload,
                                                 resolution, start_date, end_date)
        else:  # list or numeric
            num_simulation_steps = int((self.end_date - self.start_date) / self.resolution) + 1

            if isinstance(transformer_preload, (int, float)):
                transformer_preload = [transformer_preload] * num_simulation_steps

            assert type(transformer_preload) is list, msg_wrong_transformer_preload_type

            # Check if all values in list are either float or int
            assert all(isinstance(x, (float, int)) for x in transformer_preload), \
                msg_invalid_value_type

            # Check whether the length of the list alignes with the simulation period and resolution
            # If num_values don't fit simulation period and no action is wanted return error
            num_values = len(transformer_preload)

            assert num_simulation_steps <= num_values or res_data is not None or repeat, \
                msg_not_enough_data_points

            if res_data is not None:
                if type(res_data) is str:
                    try:
                        date = datetime.datetime.strptime(res_data, '%H:%M:%S')
                        res_data = datetime.timedelta(hours=date.hour, minutes=date.minute,
                                                      seconds=date.second)
                    except ValueError:
                        logging.error('%s is of incorrect format. Please use %s',
                                      res_data, '%H:%M:%S')
                        raise ValueError
                assert isinstance(res_data, datetime.timedelta), msg_wrong_resolution_type

                transformer_preload = adjust_resolution(transformer_preload, res_data,
                                                        self.resolution)

            if repeat is True:
                transformer_preload = repeat_data(transformer_preload, num_simulation_steps)

            # Should it be ==?
            assert len(transformer_preload) >= num_simulation_steps, msg_alignement_unsuccessful

        self.transformer_preload = transformer_preload

        return self

    def with_scheduling_policy(self, scheduling_policy_input):
        """Update the scheduling policy to use.
        Default: :obj: `elvis.sched.schedulers.Uncontrolled`.
        Use default if input not a str or str can not be matched.

        Args:
            scheduling_policy_input: Either str containing name of the scheduling policy to be used.
                Or instance of :obj: `elvis.sched.schedulers.SchedulingPolicy`.
        """
        # set default
        scheduling_policy = schedulers.Uncontrolled()

        # if input is already instance of Scheduling Policy assign
        if isinstance(scheduling_policy_input, schedulers.SchedulingPolicy):
            self.scheduling_policy = scheduling_policy_input
            return self

        # ensure input is str. If not return default
        if type(scheduling_policy_input) is not str:
            logging.error('Scheduling policy should be of type str or an instance of '
                          'SchedulingPolicy. The uncontrolled strategy has been used as a default.')
            self.scheduling_policy = scheduling_policy
            return self

        # Match string
        if scheduling_policy_input in ('Uncontrolled', 'UC', 'Uc', 'uc'):
            self.scheduling_policy = schedulers.Uncontrolled()

        elif scheduling_policy_input in ('Discrimination Free', 'DF', 'df'):
            self.scheduling_policy = schedulers.DiscriminationFree()

        elif scheduling_policy_input == 'FCFS':
            self.scheduling_policy = schedulers.FCFS()

        elif scheduling_policy_input in ('With Storage', 'ws', 'WS'):
            self.scheduling_policy = schedulers.WithStorage()

        elif scheduling_policy_input in ('Optimized', 'opt', 'OPT'):
            self.scheduling_policy = schedulers.Optimized()

        # invalid str use default: Uncontrolled
        else:
            logging.error('"%s" can not be matched to any existing scheduling policy.'
                          'Please use: "Uncontrolled", "Discrimination Free", "FCFS", '
                          '"With Storage" or "Optimized". '
                          'Uncontrolled is the default value and has been used for the simulation.',
                          str(scheduling_policy_input))

            self.scheduling_policy = schedulers.Uncontrolled()
        return self

    def with_df_charging_period(self, charging_period):

        assert isinstance(charging_period, (str, datetime.timedelta)), (
            'Charging period must either be a str in format: %H:%M:%S or an instance '
            'of datetime.timedelta')

        # str
        if type(charging_period) is str:
            try:
                date = datetime.datetime.strptime(charging_period, '%H:%M:%S')
                self.df_charging_period = datetime.timedelta(hours=date.hour,
                                                             minutes=date.minute,
                                                             seconds=date.second)
            except ValueError:
                try:
                    seconds = pd.Timedelta(charging_period).total_seconds()
                    self.df_charging_period = datetime.timedelta(seconds=seconds)
                except ValueError:
                    print('Incorrect timedelta format for resolution pls use: %H:%M:%S or '
                          'a pandas conform timedelta format.')
        # datetime.timedelta
        else:
            self.df_charging_period = charging_period

        return self

    def with_emissions_scenario(self, emissions_scenario, start_date, end_date, resolution,
                                col_pos=0, res_data=None, repeat=None):
        """Update emissions scenario variable representing the CO2-emissions per kWh at a given
        time.

        Args:
            emissions_scenario: Can be either of type int, float, list or pandas Series or
            DataFrame.
            start_date: (:obj: `datetime.datetime`): First time stamp.
            end_date: (:obj: `datetime.datetime`): Upper bound for time stamps.
            resolution: (:obj: `datetime.timedelta`): Time in between two adjacent time stamps.
            col_pos: (int): If pandas DataFrame is passed indicates the column in which the
                emission values are stored.
            res_data: (datetime.timedelta): If list is passed that does not align with the
                simulation period, res_data displays the period inbetween two adjacent list entries.
            repeat: (bool): If list is passed that does not align with the
                simulation period and repeat is True the list is repeated until enough data points
                are available for the simulation period.
            """
        if emissions_scenario is None:
            return self

        assert isinstance(start_date, datetime.datetime), 'Start date must be datetime.datetime.'
        assert isinstance(end_date, datetime.datetime), 'End date must be datetime.datetime.'
        assert isinstance(resolution, datetime.timedelta), 'Resolution must be datetime.timedelta'
        assert isinstance(col_pos, int), 'Column must be a positive integer.'
        assert col_pos >= 0, 'Column must be a positive integer.'

        # Error messages
        msg_wrong_resolution_type = "If the transformer preload is passed as list and the" \
                                    "resolution is supposed to be of adjusted please" \
                                    "use type datetime.timedelta."
        msg_alignment_unsuccessful = "Transformer preload could not be aligned to simulation" \
                                     "steps."
        msg_invalid_value_type = "Transformer preload should be of type: pandas DataFrame, " \
                                 "list containing float or int, int, float."
        msg_not_enough_data_points = "There are less values for the transformer preload than " \
                                     "simulation steps. Please adjust the data."

        emissions_scenario_aligned = []

        # Constant value
        if isinstance(emissions_scenario, (float, int)):
            self.emissions_scenario = [emissions_scenario] * \
                                         num_time_steps(start_date, end_date, resolution)

        # pandas DataFrame
        elif isinstance(emissions_scenario, pd.DataFrame):
            emissions_scenario = emissions_scenario.iloc[:, col_pos]

            emissions_scenario_aligned = transform_data(emissions_scenario,
                                                        resolution, start_date, end_date)
        elif isinstance(emissions_scenario, pd.Series):
            emissions_scenario_aligned = transform_data(emissions_scenario,
                                                        resolution, start_date, end_date)
        elif isinstance(emissions_scenario, list):
            num_simulation_steps = num_time_steps(start_date, end_date, resolution)

            # Check if all values in list are either float or int
            assert all(isinstance(x, (float, int)) for x in emissions_scenario), \
                msg_invalid_value_type

            # Check whether the length of the list alignes with the simulation period and resolution
            # If num_values don't fit simulation period and no action is wanted return error
            num_values = len(emissions_scenario)

            assert num_simulation_steps < num_values or res_data is not None or repeat, \
                msg_not_enough_data_points

            if res_data is not None:
                if type(res_data) is str:
                    try:
                        date = datetime.datetime.strptime(res_data, '%H:%M:%S')
                        res_data = datetime.timedelta(hours=date.hour, minutes=date.minute,
                                                      seconds=date.second)
                    except ValueError:
                        logging.error('%s is of incorrect format. Please use %s',
                                      res_data, '%H:%M:%S')
                        raise ValueError
                assert isinstance(res_data, datetime.timedelta), msg_wrong_resolution_type

                emissions_scenario = adjust_resolution(emissions_scenario, res_data,
                                                       self.resolution)

            if repeat is True:
                emissions_scenario = repeat_data(emissions_scenario, num_simulation_steps)

            # Should it be ==?
            assert len(emissions_scenario) >= num_simulation_steps, msg_alignment_unsuccessful
        else:
            warnings.warn('Emissions are passed in an non convertible type. Please either use:'
                          'Float/int, list or pandas Series or DataFrame. The simulation is carried'
                          ' on without assigning emission values.')

        return emissions_scenario_aligned

    def with_opening_hours(self, opening_hours):
        """Update the opening hours to use."""

        if opening_hours is None:
            self.opening_hours = opening_hours
            return self

        assert isinstance(opening_hours, tuple), 'Opening hours is expected to be a tuple.'
        assert len(opening_hours) == 2, 'Opening hours is expected to be a tuple with 2 values'
        _open = opening_hours[0]
        _close = opening_hours[1]
        assert isinstance(_open,
                          (float, int)), 'Values in opening hours must be of type int or ' \
                                         'float representing the hours of the day.'
        assert isinstance(_open,
                          (float, int)), 'Values in opening hours must be of type int or ' \
                                         'float representing the hours of the day.'

        assert _open <= _close, 'The first value (opening hour) is expected to be smaller than ' \
                                'the 2nd value (closing hour).'
        assert _close <= 24, 'The last value(closing hour) is expected to be smaller or equal to 24'
        assert _open >= 0, 'The first value(opening hour) is expected to be bigger or equal to 0'

        self.opening_hours = opening_hours
        return self



from numpy import histogram
from datetime import timedelta
import numpy as np
from elvis.charging_point import ChargingPoint
from elvis.config import ScenarioRealisation
from elvis.utility.elvis_general import num_time_steps, create_time_steps
from elvis.distribution import EquallySpacedInterpolatedDistribution
from elvis.infrastructure_node import Storage


class ElvisResult:
    """Represents the result of an Elvis simulation. 

    Don't use this directly, instead pass it to one of the provided functions for evaluating Elvis
    results.
    """

    def __init__(self, scenario=None, realisation_file_name=None):
        self.power_charging_points = {}
        self.power_storage_systems = {}
        self.aggregated_load_profile = None
        self.counter_rejections = 0
        self.charging_periods = None

        # used to cache the last saved timestamp for each charging point
        self._last_stored_cp_power = {}
        # used to cache the last save timestamp for the storage system
        self._last_stored_storage_power = {}

        if scenario is not None:
            assert isinstance(scenario, ScenarioRealisation), 'Result.scenario must be of type ' \
                                                              'ScenarioRealisation'
        self.scenario = scenario

        if realisation_file_name is not None:
            assert self.scenario is not None, 'To store the simulated ScenarioRealisation the ' \
                                              'scenario must be passed as type of ' \
                                              'ScenarioRealisation.'
            self.scenario.save_to_disk(r'../data/realisations/' + str(realisation_file_name))

    def store_power_charging_points(self, power_charging_points, pos_current_time_stamp, is_last_step):
        """Adds the key pos_current_time_stamp and the power assigned to the individual charging
            point to self.power_charging_points if the assigned power is not 0.

            Args:
                power_charging_points: (dict): Containing the charging points as keys as type
                    :obj: `cp.ChargingPoint`. Values are the powers to be assigned
                    as float.
                pos_current_time_stamp: (int): Position of the current time stamp in the list
                    containing all time stamps.
        """
        for cp in power_charging_points:
            assert isinstance(cp, ChargingPoint)

            power = power_charging_points[cp]
            
            if cp.id not in self.power_charging_points:
                self.power_charging_points[cp.id] = {}
            elif not is_last_step and (cp.id in self._last_stored_cp_power):
                if self._last_stored_cp_power[cp.id] == power:
                    continue
            
            self._last_stored_cp_power[cp.id] = power
            self.power_charging_points[cp.id][pos_current_time_stamp] = power

    def store_power_storage_systems(self, power_storage_systems, pos_current_time_stamp, is_last_step):
        """Saves the power assigned to the storage system in each time step.

        Args:
            power_storage_systems: (dict)): Power assigned to all storage systems.
            pos_current_time_stamp: (int): Current time step.
            is_last_step: (bool): Denotes whether the end of the simulation is reached.

        """

        for storage_system in power_storage_systems:
            assert isinstance(storage_system, Storage)

            power = power_storage_systems[storage_system]

            if storage_system.id not in self.power_storage_systems:
                self.power_storage_systems[storage_system.id] = {}
            elif not is_last_step and (storage_system.id in self._last_stored_storage_power):
                if self._last_stored_storage_power[storage_system.id] == power:
                    continue

            self._last_stored_storage_power[storage_system.id] = power
            self.power_storage_systems[storage_system.id][pos_current_time_stamp] = power

    def to_yaml(self):
        """Serialize this ElvisResult to a yaml string."""

        pass

    def to_csv(self, file_name):
        """Serialize this ElvisResult to a CSV file."""

        pass

    def aggregate_load_profile(self, num_simulation_steps=None):

        if num_simulation_steps is None:
            assert self.scenario is not None, 'If using result.aggregate_load_profile without ' \
                                              'passing the number of simulation steps the ' \
                                              'field result.scenario must be set to the scenario ' \
                                              'realisation.'

            num_simulation_steps = num_time_steps(self.scenario.start_date, self.scenario.end_date,
                                                  self.scenario.resolution)
        load_profile = []
        power = {}
        for time_step in range(num_simulation_steps):
            for cp in self.power_charging_points:
                if time_step in self.power_charging_points[cp].keys():
                    power[cp] = self.power_charging_points[cp][time_step]

            load_profile.append(sum(power.values()))

        self.aggregated_load_profile = load_profile

        return load_profile

    def get_storage_profile(self, num_simulation_steps=None):

        if num_simulation_steps is None:
            assert self.scenario is not None, 'If using result.get_storage_profile without ' \
                                              'passing the number of simulation steps the ' \
                                              'field result.scenario must be set to the scenario ' \
                                              'realisation.'

            num_simulation_steps = num_time_steps(self.scenario.start_date, self.scenario.end_date,
                                                  self.scenario.resolution)
        storage_profile = []
        power = {}
        for time_step in range(num_simulation_steps):
            for stor in self.power_storage_systems:
                if time_step in self.power_storage_systems[stor].keys():
                    power[stor] = self.power_storage_systems[stor][time_step]

            storage_profile.append(sum(power.values()))

        return storage_profile

    def total_energy_charged(self):
        power = 0
        for cp in self.power_charging_points:
            for time_step in self.power_charging_points[cp]:
                power += self.power_charging_points[cp][time_step]
        return float(power)

    def max_load(self):
        """Calculates the max load of an aggregated load profile:

            max_load: (float): Max load that is generated by the charging infrastructure.
        """
        assert self.aggregated_load_profile is not None, 'Before calculating KPIs the aggreagated' \
                                                         'load profile must be calculated.'
        return max(self.aggregated_load_profile)

    def simultaneity_factor(self, infrastructure=None, bins=None, quantile=None):
        """Calculates the simultaneity factor of the infrastructure.
            If bins is specified a histogram is returned.
            If bins is None only the max value is returned.

            The simultaneity factor for each time step is calculated via:
                (power assigned to all charging points) / (nominal power of all charging points)
            Args:
                infrastructure: (dict): Elvis conform infrastructure from which the nominal power
                    of all charging points is generated from.
                bins: (list): Containing the values for the bins as per numpy.histogram.

            returns.
                simultaneity_factor: (list or float):
                    If bins == None:
                        Float representing the max simultaneity that was reached over the simulation
                        period.
                    If bins is not None:
                        list of tuples containing the bin and the amount of time steps in which the
                        simultaneity factor is within that bin.
                """

        if (infrastructure is None) and (self.scenario is not None):
            infrastructure = self.scenario.infrastructure
        elif infrastructure is None and self.scenario is None:
            raise ValueError('If no scenario is stored in result an infrastructure has to be passed'
                             'to simultaneity factor to calculate the nominal power of the '
                             'charging points.')

        if self.aggregated_load_profile is None:
            assert self.scenario is not None, 'To calculate the simultaneity_factor the ' \
                                               'aggregated load profile must be calculated.'

            time_steps = num_time_steps(self.scenario.start_date, self.scenario.end_date,
                                        self.scenario.resolution)

            self.aggregate_load_profile(time_steps)
        assert self.aggregated_load_profile is not None, 'Before calculating KPIs the aggregated' \
                                                         'load profile must be calculated.'

        if (bins is None) and (quantile is None):
            power_hardware_max = self.get_power_charging_points(infrastructure)
            power_max = self.max_load()
            return power_max / power_hardware_max
        elif (bins is None) and not (quantile is None):
            power_hardware_max = self.get_power_charging_points(infrastructure)
            cf = np.array(self.aggregated_load_profile)/power_hardware_max
            return np.quantile(cf, q=quantile)
        else:
            power_hardware_max = self.get_power_charging_points(infrastructure)
            sim = [i / power_hardware_max for i in self.aggregated_load_profile]
            hist = histogram(sim, bins)
            return list(zip(hist[0], hist[1]))

    def electricity_costs_fix(self, electricity_rate):
        """Calculates the total electricity costs based on the electricity received from the grid
        assuming a constant electricity rate.

        Args:
            electricity_rate: (float): Costs per kWh.

        Returns:
            electricity_costs: (float): Total costs of electricity.

        TODO:
            - Storage integration
            - PV integration
        """

        total_energy = self.total_energy_charged()
        electricity_costs = total_energy * electricity_rate

        return electricity_costs

    def electricity_costs_24_variable(self, variable_electricity_rate):
        """Calculates the total electricity costs based on the electricity receives from the grid
        assuming a variable electricity during a day. This intraday cost distribution is assumed to
        be same for every day of the simulation.

        Args:
            variable_electricity_rate: (list): Containing floats representing the costs of
            electricity during one day.

        Returns:
            electricity_costs: (float): Total costs of electricity.

        TODO:
            - Storage integration
            - PV integration
        """

        assert type(variable_electricity_rate) == list, 'The variable electricity rate must be of '\
                                                        'type list.'

        msg_invalid_value_type = "Arrival distribution should be of type: pandas DataFrame or a " \
                                 "list containing float or int."

        # Check that all values in list are either float or int
        assert all(isinstance(x, (float, int)) for x in variable_electricity_rate), \
            msg_invalid_value_type

        time_coeff = 24/len(variable_electricity_rate)
        hour_stamps_costs = [x * time_coeff for x in range(len(variable_electricity_rate))]

        # Repeat first cost value
        hour_stamps_costs.append(24 + hour_stamps_costs[0])
        variable_electricity_rate.append(variable_electricity_rate[0])
        cost_distr = EquallySpacedInterpolatedDistribution.linear(
            list(zip(hour_stamps_costs, variable_electricity_rate)), None)

        time_stamps = create_time_steps(self.scenario.start_date, self.scenario.end_date,
                                        self.scenario.resolution)

        load_profile = self.aggregate_load_profile()

        electricity_costs = 0
        for i in list(range(len(load_profile))):
            p = load_profile[i]
            # get time stamp
            current_time_stamp = time_stamps[i]
            # calc the time in hours
            seconds = current_time_stamp.hour * 3600 + current_time_stamp.minute * 60 + \
                      current_time_stamp.second + current_time_stamp.microsecond / 1000000
            x_pos_time_stamp = seconds / 3600
            # lookup the price at the current time (linearly interpolated)
            price = cost_distr[x_pos_time_stamp]
            # add costs
            electricity_costs += p * price

        return electricity_costs

    def total_emissions(self):
        assert self.scenario is not None, 'Scenario in result must be specified in order to ' \
                                          'calculate the emissions.'
        assert self.scenario.emissions_scenario is not None, 'There must be a emissions_scenario ' \
                                                             'specified in order to calculate ' \
                                                             'emissions.'

        load_profile = self.aggregate_load_profile()
        emissions = self.scenario.emissions_scenario

        total_emissions = 0
        for i in range(len(load_profile)):
            total_emissions += load_profile[i] * emissions[i]

        return total_emissions

    def average_charging_time(self, in_hours=False):
        """Calculated the average parking time.
        Definition: The average time between the first time the car is charged and the last time.

        Args:
            in_hours: (bool): If true return average charging time as hours (float).
        Returns:
            average_charging_time: (datetime.timedelta): Mean time needed to charge a car.
        """

        assert self.charging_periods is not None, 'Charging periods must be assigned.'
        assert isinstance(self.charging_periods, dict), 'Charging periods should be of type dict.'

        event_counter = 0
        seconds = 0
        for event in self.charging_periods:
            event_counter += 1
            end = self.charging_periods[event]['last_charged']
            start = self.charging_periods[event]['arrival']
            seconds += (end - start).total_seconds()

        if in_hours is True:
            return seconds/3600/event_counter

        return timedelta(seconds=seconds/event_counter)


    @staticmethod
    def from_yaml(yaml_str):
        """Create an ElvisResult from a yaml string."""

        pass

    @staticmethod
    def from_json(json_str):
        """Create an ElvisResult from a yaml string."""

        pass

    @staticmethod
    def from_csv(file_name):
        """Create an ElvisResult from a CSV file."""

        pass

    @staticmethod
    def get_power_charging_points(infrastructure):
        """Sums over the power of all charging points in an Elvis conform infrastructure dict.

        Args:
            infrastructure: Elvis conform infrastructure dict.

        Returns:
            power_sum: (float): Sum over the power of all charging points.
        """
        msg_wrong_infrastructure = "Infrastructure must be Elvis conform."
        assert isinstance(infrastructure, dict), msg_wrong_infrastructure
        assert 'transformers' in infrastructure.keys(), msg_wrong_infrastructure
        power_sum = 0
        for transformer in infrastructure['transformers']:
            assert 'charging_stations' in transformer.keys(), msg_wrong_infrastructure
            for charging_station in transformer['charging_stations']:
                assert 'charging_points' in charging_station.keys(), msg_wrong_infrastructure
                for cp in charging_station['charging_points']:
                    assert 'max_power' in cp.keys()
                    assert isinstance(cp['max_power'], (float, int))
                    power_sum += cp['max_power']

        return power_sum



"""Wrapper class to have a more simplistic object than queue.WaitingQueue.

TODO: Question: Is setting self.next_leave to year 9999 okay? Does the job but looks weird.
-> used to not have to check if len==0 or next_leave==None. see simulate.py, line: 89"""

import datetime


class WaitingQueue:
    """Represents the waiting queue for vehicles."""
    def __init__(self, maxsize=0, arrivals=None):

        self.queue = []
        # initialise with large enough date
        self.next_leave = datetime.datetime(9999, 1, 1)
        if arrivals is not None:
            for arrival in arrivals:
                self.queue.append(arrival)

            self.determine_next_leaving_time()
        self.maxsize = maxsize

    def enqueue(self, item):
        """Add item to queue if not already full."""
        if self.size() < self.maxsize:
            self.queue.append(item)
            self.determine_next_leaving_time()

    def dequeue(self):
        """Get next item from queue if not empty."""
        if len(self.queue) < 1:
            return None
        self.determine_next_leaving_time()
        return self.queue.pop(0)

    def size(self):
        """Get number of item in queue."""
        return len(self.queue)

    def determine_next_leaving_time(self):
        """Determine the next departure time of all queued vehicles."""
        if self.size() > 0:
            leaving_times = []
            for event in self.queue:
                arrival_time = event.arrival_time
                parking_time = datetime.timedelta(hours=event.parking_time)

                leaving_times.append(arrival_time + parking_time)
            self.next_leave = sorted(leaving_times)[0]
        else:
            # if len = 0: Set next leave to be large enough date
            self.next_leave = datetime.datetime(9999, 1, 1)

    def empty(self):
        """Disconnect all vehicles currently enqueued"""
        self.queue = []




class Power:
    """Unit for electrical power."""

    def __init__(self, watts):
        self.watts = watts

    @property
    def kilowatts(self):
        return self.watts / 1000.0

class Current:
    """Unit for electrical current."""

    def __init__(self, amps):
        self.amps = amps

    @property
    def milliamps(self):
        return self.amps * 1000.0

    def __mul__(self, hours):
        return Charge(self.amps * hours)

class Charge:
    """Unit for electrical charge."""

    def __init__(self, amp_hours):
        self.amp_hours = amp_hours

    def energy(self, voltage):
        return Energy(amp_hours * voltage)

class Energy:
    """Unit for electrical energy."""

    def __init__(self, kwh):
        self.kwh = kwh

    def charge(self, voltage):
        return Charge(kwh / voltage)



from elvis.utility.elvis_general import floor
from elvis.units import Energy
from elvis.units import Power


class Battery:
    """Models a generic battery."""

    def __init__(self, capacity: Energy, max_charge_power: Power,
                 min_charge_power: Power, efficiency: float, start_power_degradation: float=1,
                 max_degradation_level: float=0):
        """Create instance of Battery given all parameters."""
        assert isinstance(capacity, (float, int)) and capacity > 0, \
            'Battery capacity must be greater than 0'
        assert isinstance(max_charge_power, (float, int))
        assert isinstance(min_charge_power, (float, int))
        assert isinstance(efficiency, (float, int)) and (0 <= efficiency <= 1)
        assert isinstance(start_power_degradation, (float, int)) and (0 <= start_power_degradation <= 1)
        assert isinstance(max_degradation_level, (float, int)) and (0 <= max_degradation_level <= 1)
        assert max_degradation_level * max_charge_power >= min_charge_power

        # battery capacity (kWh)
        self.capacity = capacity

        # maximum supported charging power (W)
        self.max_charge_power = max_charge_power

        # minimum supported charging power (W)
        self.min_charge_power = min_charge_power

        # battery efficiency ([0,1])
        self.efficiency = efficiency

        # SOC level at which the max power start degrading ([0,1])
        self.start_power_degradation = start_power_degradation

        # Remaining fraction of initial max_power at SOC = 1 (max degradation) ([0,1])
        # smaller max_degradation_level -> power degrades more:
        # max_power_possible * max_degradation_level
        self.max_degradation_level = max_degradation_level

    def to_dict(self):
        dictionary = self.__dict__.copy()
        return dictionary

    def max_power_possible(self, current_soc):
        """Return the max power that can be assigned to the battery in regard of the current
            SOC.
            As of implementation right now return power assignable to the car:
                Max_power_at_battery_pins / conversion_efficiency.

        Args:
            current_soc: (float): Current state of charge: [0, 1]

        Return:
            max_power_possible: (float): Max assignable power.

        """
        if current_soc > self.start_power_degradation:

            power_degradation = (current_soc - self.start_power_degradation) / \
                                (1-self.start_power_degradation) * \
                                self.max_charge_power * (1 - self.max_degradation_level)

            max_power_possible = self.max_charge_power - power_degradation
            return max_power_possible

        return self.max_charge_power

    def min_power_possible(self, current_soc):
        """Return the min power that can be assigned to the battery in regard of the current
        SOC.

        Args:
            current_soc: (float): Current state of charge: [0, 1]

        Return:
            max_power_possible: (float): Max assignable power.

        TODO: Implement SOC dependence."""

        return self.min_charge_power

    @staticmethod
    def from_dict(**kwargs):
        """Initialise an instance of ChargingEvent with values stored in a dict.

        Args:

            **kwargs: Arbitrary keyword arguments.
        """

        necessary_keys = ['capacity', 'max_charge_power', 'min_charge_power', 'efficiency']

        for key in necessary_keys:
            assert key in kwargs, 'Not all necessary keys are included to create an EVBattery ' \
                                  'from dict. Missing: ' + key

        capacity = kwargs['capacity']
        max_charge_power = kwargs['max_charge_power']
        min_charge_power = kwargs['min_charge_power']

        efficiency = kwargs['efficiency']

        if 'start_power_degradation' and 'max_degradation_level' in kwargs:
            start_power_degradation = kwargs['start_power_degradation']
            max_degradation_level = kwargs['max_degradation_level']

            return EVBattery(capacity, max_charge_power, min_charge_power, efficiency,
                             start_power_degradation, max_degradation_level)

        return EVBattery(capacity, max_charge_power, min_charge_power, efficiency)


class EVBattery(Battery):
    """Models an electric vehicle battery."""

    def __init__(self, *args, **kwargs):
        super(EVBattery, self).__init__(*args, **kwargs)


class StationaryBattery(Battery):
    """Models a stationary vehicle battery."""

    def __init__(self, *args, **kwargs):

        keys = kwargs.keys()

        if 'initial_soc' in keys:
            assert isinstance(kwargs['initial_soc'], (int, float)), 'Initial SOC must be numeric.'
            assert 0 <= kwargs['initial_soc'] <= 1, 'Initial SOC must be in between 0 and 1'
            self.soc = kwargs['initial_soc']
        else:
            self.soc = 0

        if 'min_soc' in keys:
            assert isinstance(kwargs['min_soc'], (int, float)), 'Min SOC must be numeric.'
            assert 0 <= kwargs['min_soc'] <= 1, 'Min SOC must be in between 0 and 1'
            self.min_soc = kwargs['min_soc']
        else:
            self.min_soc = 0

        self.power = 0
        self.soc_time = []

        super(StationaryBattery, self).__init__(*args, **kwargs)

    def max_discharge_power(self, cur_ass_power, step_length):
        """Calculates the max power possible to discharge the storage system regarding its power
            limits and its current energy level.

        Args:
            cur_ass_power: (float): Currently assigned power.
            step_length: (:obj: `datetime.timedelta`): Resolution of the simulation denoting the
                time in between two adjacent time steps.

        Returns:
            max_power: (float): Max power possible to discharge the storage system with.

        TODO:
            - integrate efficiencies
        """
        assert isinstance(cur_ass_power, (float, int)), 'Currently assigned power should be numeric'
        # Max power based on already assigned power and what is theoretically possible to discharge
        max_power_theo = self.max_charge_power - cur_ass_power

        # Power that leads to SOC = min_soc at the end of time step
        step_length_hours = step_length.total_seconds() / 3600
        soc_cur_ass_power = cur_ass_power / self.capacity * step_length_hours
        power_to_empty = (self.soc - soc_cur_ass_power - self.min_soc) * \
                         self.capacity / step_length_hours

        # Max power possible based on energy level and power limits
        max_power = max(min(max_power_theo, power_to_empty), 0)
        # Round down to 3 decimals
        max_power = floor(max_power)
        return max_power

    def charge(self, available_power, step_length):
        """
        Charges the battery with a given power or with whatever is possible considering its
            power limits and its current energy level.

        Args:
            available_power: (float): Power that is available from the transformer.
            step_length: (:obj: `datetime.timedelta`): Resolution of the simulation denoting the
                time in between two adjacent time steps.

        Returns:
            power_charged: The power the battery was actually charged with.

        TODO:
            - integrate efficiencies
        """
        assert isinstance(available_power, (float, int)), 'Available power should be numeric'
        assert available_power >= 0, 'The max charge power must be >= 0.'
        # Power needed to charge to SOC = 1 at the end of time step
        step_length_hours = step_length.total_seconds() / 3600
        max_power_to_full = (1 - self.soc) * self.capacity / step_length_hours

        # Max power possible considering all limits (available, power limits, energy level)
        power_charged = min(max_power_to_full, self.max_charge_power, available_power)

        # SOC change given the power to charge with
        delta_soc = power_charged * step_length_hours / self.capacity
        self.soc += delta_soc
        # Make sure SOC is within limits
        self.check_soc()
        self.power = power_charged
        self.soc_time.append(self.soc)
        return power_charged

    def discharge(self, power_to_discharge, step_length):
        """
        Tries to discharge the battery with given power. If either the power limits or the limit
            due to the current energy level is violated a ValueError will be raised.

        Args:
            power_to_discharge: (float): Power the system is supposed to be discharged with.
                Only positive values are allowed. These are understood to be discharged though.
            step_length: (:obj: `datetime.timedelta`): Resolution of the simulation denoting the
                time in between two adjacent time steps.

        TODO:
            - integrate efficiencies
            """
        assert isinstance(power_to_discharge,(float, int)), 'Currently assigned power ' \
                                                            'should be numeric'
        assert power_to_discharge >= 0, 'The max charge power must be >= 0.'

        # Max power too discharge with at current SOC
        max_discharge_power = self.max_discharge_power(0, step_length)

        if max_discharge_power < power_to_discharge:
            raise ValueError('Power to discharge with is out of the limits.')
        self.power = - power_to_discharge

        # SOC change given the power to discharge with
        step_length_hours = step_length.total_seconds() / 3600
        delta_soc = power_to_discharge * step_length_hours / self.capacity
        self.soc -= delta_soc
        # Make sure SOC is within limits
        self.check_soc()
        self.soc_time.append(self.soc)
        return

    def check_soc(self):
        assert 0 <= self.soc <= 1, 'SOC is out of limits.'



"""Class representing vehicle types. Especially in regard of their battery."""
from elvis.battery import EVBattery


class ElectricVehicle:
    """Models the charging behaviour of a specific EV model."""

    def __init__(self, brand: str, model: str, battery: EVBattery, probability):
        assert isinstance(battery, EVBattery)
        assert isinstance(probability, (float, int))

        self.brand = brand
        self.model = model
        self.battery = battery
        self.probability = probability

    def __str__(self):
        printout = self.brand + ', ' + self.model
        return printout

    def to_dict(self):
        dictionary = self.__dict__.copy()
        dictionary['battery'] = self.battery.to_dict()
        return dictionary

    @staticmethod
    def from_dict(**kwargs):
        """Initialise an instance of ChargingEvent with values stored in a dict.

        Args:

            **kwargs: Arbitrary keyword arguments.
        """

        necessary_keys = ['brand', 'model', 'battery', 'probability']

        for key in necessary_keys:
            assert key in kwargs, 'Not all necessary keys are included to create a VehicleType ' \
                                  'from dict.'

        brand = kwargs['brand']
        model = kwargs['model']
        probability = kwargs['probability']
        battery = EVBattery.from_dict(**kwargs['battery'])

        return ElectricVehicle(brand, model, battery, probability)




import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    with open('requirements.txt') as req:
        REQUIREMENTS = [r.partition('#')[0] for r in req if not r.startswith('-e')]
except OSError:
    print("Error when reading requirements.txt")
    REQUIREMENTS = []

setuptools.setup(
    name="py-elvis",
    version="0.2.0",
    author="Moritz Markschläger, Jonas Zell, Marcus Voss, Izgh Hadachi",
    author_email="moritz.markschlaeger@dai-labor.de",
    description="A planning and management tool for electric vehicles charging infrastructure",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)